import numpy as np
from waveforms.dicttree import _eq
from waveforms.registry import DELETE, Registry, TreeRef
from config import config


def test_registry():
    reg = Registry.from_url('sqlite:///:memory:')
    assert reg.export() == {}

    reg.set('test.a.b', 12.5)
    assert reg.export() == {'test': {'a': {'b': 12.5}}}

    reg.update({'test': {'x': 'hello'}})
    assert reg.export() == {'test': {'x': 'hello', 'a': {'b': 12.5}}}

    reg.update(
        {'dev': {
            'a': '12',
            'b': 'tcp',
            'c': 11283,
            'd': np.arange(10)
        }})

    assert np.equal(reg.get('dev.d[::2]'), np.array([0, 2, 4, 6, 8])).all()
    assert _eq(
        reg.export(), {
            'test': {
                'x': 'hello',
                'a': {
                    'b': 12.5
                }
            },
            'dev': {
                'a': '12',
                'b': 'tcp',
                'c': 11283,
                'd': np.arange(10)
            }
        })

    reg.update({'dev': {'b': {'c': 12}, 'c': DELETE}})
    assert _eq(
        reg.export(), {
            'test': {
                'x': 'hello',
                'a': {
                    'b': 12.5
                }
            },
            'dev': {
                'a': '12',
                'd': np.arange(10),
                'b': {
                    'c': 12
                }
            }
        })

    assert reg.get('dev.c') is None
    assert reg.previous().get('dev.c') == 11283
    assert _eq(reg.get('dev.d'), np.arange(10))
    assert _eq(reg.previous().get('dev.d'), np.arange(10))


def test_search():
    reg = Registry.from_url('sqlite:///:memory:')
    reg.update(config)

    ret = reg.export(depth=1)

    assert set(ret.keys()) == {'__version__', 'station', 'chip', 'gates'}
    assert all(
        isinstance(ret[k], TreeRef) for k in ['station', 'chip', 'gates'])
    assert ret['__version__'] == 1

    assert reg.get('gates.rfUnitary.Q1.params.amp[1][1]') == 0.8204
    assert list(reg.search('gates.rfUnitary.*.params.amp[1][1]')) == [
        ('gates.rfUnitary.Q1.params.amp[1][1]', 0.8204),
        ('gates.rfUnitary.Q2.params.amp[1][1]', 0.658)
    ]

    for k, v in reg.search('gates.!rfUnitary.*.params'):
        assert k.startswith('gates.') and k.endswith('.params')
        assert not k.startswith('gates.rfUnitary')
        assert isinstance(v, TreeRef)

    for k, v in reg.search('gates.*.*.*.dur*'):
        assert k.endswith('.duration')
        assert isinstance(v, (list, float))

    for k, v in reg.search('gates.*.*.*.dur*[1][1]'):
        assert k.startswith('gates.rfUnitary.Q') and k.endswith(
            '.params.duration[1][1]')
        assert isinstance(v, float)



import numpy as np
from scipy.optimize import leastsq

from waveforms.math.fit.simple import fit_circle
from waveforms.math.signal.func import complexPeaks


def get_unit_prefix(value):
    '''
    获取 value 合适的单位前缀，以及相应的倍数

    >>>
        y => 1e-24       Y => 1e24
        z => 1e-21       Z => 1e21
        a => 1e-18       E => 1e18
        f => 1e-15       P => 1e15
        p => 1e-12       T => 1e12
        n => 1e-9        G => 1e9
        u => 1e-6        M => 1e6
        m => 1e-3        k => 1e3

    Returns:
        (prefix, multiple)
    '''
    prefixs = [
        'y', 'z', 'a', 'f', 'p', 'n', 'u', 'm', '', 'k', 'M', 'G', 'T', 'P',
        'E', 'Z', 'Y'
    ]
    if value == 0:
        return '', 1
    x = np.floor(np.log10(abs(value)) / 3)
    x = 0 if x < -8 else x
    x = 0 if x > 8 else x
    return prefixs[int(x) + 8], 1000**(x)


def valueString(value, unit=""):
    """
    将 value 转换为更易读的形式

    >>> valueString(1.243e-7, 's')
    ... "124.3 ns"
    >>> valueString(1.243e10, 'Hz')
    ... "12.43 GHz"
    """
    prefix, k = get_unit_prefix(value)
    return f"{value/k:g} {prefix}{unit}"


def S21(f, f0, QL, Qc, phi):
    width = f0 / (2 * QL)
    amp = -QL / np.abs(Qc) * np.exp(1j * phi)
    return complexPeaks(f, [(f0, width, amp)], 1)


def invS21(freq, f0, Qc, Qi, phi):
    width = f0 / (2 * Qi)
    amp = Qi / np.abs(Qc) * np.exp(1j * phi)
    return complexPeaks(freq, [(f0, width, amp)], 1)


def getBackground(x, s):
    phi = np.unwrap(np.angle(s), 0.9 * np.pi)
    delay = (phi[-1] - phi[0]) / (2 * np.pi * (x[-1] - x[0]))
    A = np.abs(s).max()

    return delay, A


def guessParams(x, s, inverse=True):
    if inverse:
        data = 1 / s
    else:
        data = s

    xc, yc, R = fit_circle(np.real(data), np.imag(data))
    y = (data - xc - 1j * yc) / R
    theta = np.angle(0.5 * (y[0] + y[-1]))
    y /= np.exp(1j * theta)
    index = np.abs(np.outer(y, np.ones(3)) -
                   np.array([-1j, -1, 1j])).argmin(axis=0)
    f1, fc, f2 = x[index]
    FWHM = abs(f2 - f1)

    if inverse:
        #f0 = x[np.abs(data).argmax()]
        Qi = fc / FWHM
        Qe = Qi / (2 * R)
        QL = 1 / (1 / Qi + 1 / Qe)
        phi = np.arctan2(yc, xc - 1)
    else:
        #f0 = x[np.abs(data).argmin()]
        QL = fc / FWHM
        Qe = QL / (2 * R)
        Qi = 1 / (1 / QL - 1 / Qe)
        phi = np.arctan2(-yc, 1 - xc)
    return [fc, Qi, Qe, QL, phi]


def background(f, params):
    f0, A, delay, Aphi, a, b = params
    return A * ((a * (f - f0)**2 + b *
                 (f - f0)) + 1) * np.exp(1j * (delay * (f - f0) + Aphi))


def fitS21(x, s, params):

    def err(params, f, s21):
        f0, QL, Qe, phi, A, Aphi, delay = params
        y = s21 - S21(f, f0, abs(QL), abs(Qe), phi) * A * np.exp(
            1j * (delay * (f - x[0]) + Aphi))
        return np.abs(y)

    f0, Qi, Qe, QL, phi = params
    params = [f0, QL, Qe, phi, 1, 0, 0]

    res = leastsq(err, params, (x, s))
    f0, QL, Qe, phi, A, Aphi, delay = res[0]
    Qi = 1 / (1 / abs(QL) - 1 / abs(Qe))
    return f0, Qi, abs(Qe), abs(QL), phi, A, Aphi, delay


def fitInvS21(x, s, params, background=(1, 0, 0)):
    A, Aphi, delay = background

    def err(params, f, s21):
        f0, QL, Qe, phi, *_ = params
        y = 1 / s21 - invS21(f, f0, abs(Qe), abs(Qi),
                             phi) / (A * np.exp(1j * (delay *
                                                      (f - x[0]) + Aphi)))
        return np.abs(y)

    f0, Qi, Qe, QL, phi = params
    params = [f0, Qe, Qi, phi]

    res = leastsq(err, params, (x, s))
    f0, Qe, Qi, phi, *_ = res[0]
    QL = 1 / (1 / abs(Qi) + 1 / abs(Qe))
    return f0, abs(Qi), abs(Qe), QL, phi, A, Aphi, delay


def plotData(fig, x, y):
    ax = fig.add_subplot(211)
    ax.plot(x / 1e9, 20 * np.log10(np.abs(y)))
    ax.set_xlabel('Frequency / GHz')
    ax.set_ylabel('|S21| / dB')

    ax = ax.twinx()
    phase = np.unwrap(np.angle(y), 0.9 * np.pi)
    ax.plot(x / 1e9, phase / np.pi, color="C1")
    ax.set_xlabel('Frequency / GHz')
    ax.set_ylabel(r'Ang(S21) / $\pi$')

    ax = fig.add_subplot(212)
    ax.plot(x / 1e9, np.real(y), label='Re')
    ax.plot(x / 1e9, np.imag(y), label='Im')
    ax.set_xlabel('Frequency / GHz')
    ax.set_ylabel('S21')
    ax.legend()


def printReport(params, method='UCSB'):

    f0, Qi, Qe, QL, phi, *others = params
    if len(others) >= 3:
        A, Aphi, delay, *others = others
    else:
        A, Aphi, delay = 1, 0, 0
        others = ()
    if len(others) == 3:
        a, b = others
    else:
        a, b = 0, 0

    print(f'''{method} Method
    f0 = {valueString(f0, 'Hz')}
    Qi = {valueString(Qi)}    Qe = {valueString(Qe)}    QL = {valueString(QL)}
    phi = {phi*180/np.pi:g} deg
    A = {valueString(A)}    delay = {valueString(delay, 's')}
    
    Aphi = {Aphi*180/np.pi:g} deg   a = {a:g} Hz^-2    b = {b:g} Hz^-1''')
    print()


def plotFit(fig, f, s, params):
    from matplotlib.gridspec import GridSpec

    f0, Qi, Qe, QL, phi, *others = params
    if len(others) == 0:
        A, Aphi, delay = 1, 0, 0
    else:
        A, Aphi, delay = others

    y = s / (A * np.exp(1j * (delay * (f - f[0]) + Aphi)))
    invY = 1 / y
    
    invs21 = invS21(f, f0, Qe, Qi, phi)
    s21 = 1 / invs21

    gs = GridSpec(4, 4, figure=fig)

    ax = fig.add_subplot(gs[0:2, 0:2])
    ax.plot(np.real(y), np.imag(y), '.')
    ax.plot(np.real(s21), np.imag(s21))
    ax.axis('equal')
    ax.set_xlabel("Re(S21)")
    ax.set_ylabel("Im(S21)")

    ax = fig.add_subplot(gs[0:2, 2:4])
    ax.plot(np.real(invY), np.imag(invY), '.')
    ax.plot(np.real(invs21), np.imag(invs21))
    ax.axis('equal')
    ax.set_xlabel("Re(S21$^{-1}$)")
    ax.set_ylabel("Im(S21$^{-1}$)")

    ax = fig.add_subplot(gs[2:3, 0:4])
    ax.plot(f / 1e9, np.real(y), '.')
    ax.plot(f / 1e9, np.imag(y), '.')
    ax.plot(f / 1e9, np.real(s21), label='Re')
    ax.plot(f / 1e9, np.imag(s21), label='Im')
    ax.set_xlabel('Frequency / GHz')
    ax.set_ylabel('S21')
    ax.legend()

    ax = fig.add_subplot(gs[3:4, 0:4])
    ax.plot(f / 1e9, 20 * np.log10(np.abs(y)), '.', color="C0")
    ax.plot(f / 1e9, 20 * np.log10(np.abs(s21)), color="C2")
    ax.set_xlabel('Frequency / GHz')
    ax.set_ylabel('|S21| / dB')

    ax = ax.twinx()
    phase = np.unwrap(np.angle(y), 0.9 * np.pi)
    ax.plot(f / 1e9, phase / np.pi, '.', color="C1", alpha=0.5)
    phase = np.unwrap(np.angle(s21), 0.9 * np.pi)
    ax.plot(f / 1e9, phase / np.pi, color="C3", alpha=0.5)
    ax.set_xlabel('Frequency / GHz')
    ax.set_ylabel(r'Ang(S21) / $\pi$')



class FibNode():
    __slots__ = ('key', 'degree', 'left', 'right', 'child', 'parent', 'marked')

    def __init__(self,
                 key=None,
                 degree=0,
                 left=None,
                 right=None,
                 child=None,
                 parent=None,
                 marked=False):
        self.key, self.degree, self.left, self.right, self.child, self.parent, self.marked = (
            key, degree, left, right, child, parent, marked)

    def add(self, node):
        self.left.right = node
        node.left = self.left
        self.left = node
        node.right = self
        node.parent = self.parent

    def cat(self, node):
        self.right.left, node.right.left = node, self
        self.right, node.right = node.right, self.right

    def remove(self):
        self.left.right, self.right.left = self.right, self.left
        self.left, self.right = self, self

    def remove_child(self, node):
        self.degree -= 1
        self.child = node.right
        node.remove()
        if self.degree == 0:
            self.child = None
        node.parent = None

    def link(self, node):
        node.remove()
        self.degree += 1
        if self.child is None:
            self.child = node
            node.left, node.right = node, node
        else:
            self.child.add(node)
            self.child = node
        node.parent = self
        node.marked = False

    def view(self):
        print(
            f'\t{self.parent}\t\n{self.left}\t{self.key}\t{self.right}\n\t{self.child}({self.degree})'
        )

    def __repr__(self):
        return f'{self.key}'  #' at {hex(id(self))}'


class FibHeap():
    """
    Fibonacci Heap

    Attributes:
        key_number: number of keys in the heap
        min_node: the node with the minimum key

    Methods:
        insert(node): insert a node into the heap
        extract_min(): extract the node with the minimum key
        consolidate(): consolidate the heap
    """

    def __init__(self, key_number=0, min_node=None):
        self.key_number, self.min_node = key_number, min_node

    def insert(self, node):
        if self.min_node is None:
            self.min_node = node
            node.left, node.right = node, node
        else:
            self.min_node.add(node)
            if node.key < self.min_node.key:
                self.min_node = node
        self.key_number += 1

    def extract_min(self):
        ret = self.min_node
        if ret is not None:
            if ret.degree:
                x = ret.child
                while ret.degree:
                    x.parent = None
                    x = x.right
                    ret.degree -= 1
                ret.cat(ret.child)
                ret.child = None
            right = ret.right
            ret.remove()
            self.key_number -= 1
            if ret == right:
                self.min_node = None
            else:
                self.min_node = right
                self.consolidate()
        return ret

    def consolidate(self):
        if self.min_node is None:
            return
        from math import log2
        max_degree = round(log2(self.key_number) + 0.5) + 1
        cons = [None] * max_degree

        root_list = [self.min_node]
        right = self.min_node.right
        while right != self.min_node:
            root_list.append(right)
            right = right.right

        for i in range(len(root_list)):
            x = root_list[i]
            while cons[x.degree] is not None:
                y = cons[x.degree]
                if y.key < x.key:
                    x, y = y, x
                x.link(y)
                cons[y.degree] = None
            cons[x.degree] = x

        self.min_node = None
        for i in range(max_degree):
            if cons[i] is not None:
                if self.min_node is None:
                    self.min_node = cons[i]
                else:
                    if cons[i].key < self.min_node.key:
                        self.min_node = cons[i]

    def cut(self, node, parent):
        parent.remove_child(node)
        self.insert(node)
        node.marked = False

    def cascade(self, node):
        parent = node.parent
        if parent is not None:
            if node.marked:
                self.cut(node, parent)
                self.cascade(parent)
            else:
                node.marked = True

    def decrease(self, node, key):
        node.key = key
        parent = node.parent
        if parent is not None and node.key < parent.key:
            self.cut(node, parent)
            self.cascade(parent)
        if node.key < self.min_node.key:
            self.min_node = node

    def increase(self, node, key):
        while node.child is not None:
            self.cut(node.child, node)
        node.degree = 0
        node.key = key
        parent = node.parent
        if parent is not None:
            self.cut(node, parent)
            self.cascade(parent)
        elif self.min_node == node:
            right = node.right
            while right != node:
                if right.key < self.min_node.key:
                    self.min_node = right
                right = right.right

    def change(self, node, key):
        if key < node.key:
            self.decrease(node, key)
        elif key > node.key:
            self.increase(node, key)

    def remove(self, node):
        self.decrease(node, self.min_node.key)
        self.min_node = node
        self.extract_min()


def heap_union(a, b):
    """
    Union two Fibonacci Heaps
    """
    if a is None or a.min_node is None:
        return b
    if b is None or b.min_node is None:
        return a
    if b.max_degree > a.max_degree:
        a, b = b, a
    a.min_node.cat(b.min_node)
    if b.min_node.key < a.min_node.key:
        a.min_node = b.min_node
    a.key_number += b.key_number
    return a


def heap_view(nodes):
    from treelib import Node, Tree

    ret = Tree()
    ret.add_node(Node(identifier='root'))
    while True:
        flag = True
        for node in nodes:
            if ret.contains(node.__repr__()):
                continue
            parent = node.parent.__repr__(
            ) if node.parent is not None else 'root'
            if ret.contains(parent):
                ret.add_node(Node(identifier=node.__repr__()), parent=parent)
            else:
                flag = False
        if flag:
            break
    return ret.show()


import inspect

from numpy import pi

from ..dicttree import NOTSET
from ..waveform import Waveform, cos, sin, step
from .base import (ADChannel, AWGChannel, Context, GateConfig, MeasurementTask,
                   MultADChannel, MultAWGChannel, QLispCode, QLispError,
                   create_context, gateName)
from .library import Library


def _ctx_update_biases(sub_ctx: Context, ctx: Context):
    for channel, bias in sub_ctx.biases.items():
        if isinstance(bias, tuple):
            bias, edge, buffer = bias
        else:
            edge, buffer = 0, 0
        if ctx.biases[channel] != bias:
            _, *qubits = channel
            t = max(ctx.time[q] for q in qubits)
            wav = (bias - ctx.biases[channel]) * step(edge) >> (t + buffer / 2)
            _addWaveforms(ctx, channel, wav)
            ctx.biases[channel] = bias


def _ctx_update_time(sub_ctx: Context, ctx: Context):
    ctx.time.update(sub_ctx.time)


def _ctx_update_phases(sub_ctx: Context, ctx: Context):
    #ctx.phases.update(sub_ctx.phases)
    for k, v in sub_ctx.phases_ext.items():
        ctx.phases_ext[k].update(v)


def _ctx_update_waveforms(sub_ctx: Context, ctx: Context):
    for channel, wav in sub_ctx.raw_waveforms.items():
        _addWaveforms(ctx, channel, wav)


def _ctx_update_measurement_tasks(sub_ctx: Context, ctx: Context):
    for cbit, task in sub_ctx.measures.items():
        hardware = ctx.cfg._getADChannel(task.qubit)
        ctx.measures[cbit] = MeasurementTask(task.qubit, task.cbit, task.time,
                                             task.signal, task.params,
                                             hardware)


def _execute(ctx, cmd):
    from waveforms.waveform import NDIGITS

    (op, target, *values), key = cmd
    if op == '!nop':
        return
    if (op, target) == ('!set', 'time'):
        ctx.time[key] = round(values[0], NDIGITS)
    elif (op, target) == ('!set', 'phase'):
        ctx.phases[key] = values[0]
    elif (op, target) == ('!set', 'phase_ext'):
        ctx.phases_ext[key][values[0]] = values[1]
    elif (op, target) == ('!set', 'bias'):
        ctx.biases[key] = values[0]
    elif (op, target) == ('!set', 'waveform'):
        ctx.raw_waveforms[key] = values[0]
    elif (op, target) == ('!set', 'cbit'):
        ctx.measures[key] = values[0]
    elif (op, target) == ('!add', 'time'):
        ctx.time[key] = round(ctx.time[key] + values[0], NDIGITS)
    elif (op, target) == ('!add', 'phase'):
        ctx.phases[key] += values[0]
    elif (op, target) == ('!add', 'phase_ext'):
        ctx.phases_ext[key][values[0]] += values[1]
    elif (op, target) == ('!add', 'bias'):
        ctx.biases[key] += values[0]
    elif (op, target) == ('!add', 'waveform'):
        ctx.raw_waveforms[key] += values[0]
    else:
        raise QLispError(f'Unknown command: {cmd}')


def _call_func_with_kwds(func, kwds):
    sig = inspect.signature(func)
    for p in sig.parameters.values():
        if p.kind == p.VAR_KEYWORD:
            return func(**kwds)
    kw = {k: v for k, v in kwds.items() if k in sig.parameters}
    return func(**kw)


def _try_to_call(x, kwds):
    if callable(x):
        return _call_func_with_kwds(x, kwds)
    return x


def _try_to_lookup_config(cfg, key):
    if isinstance(key, str) and key.startswith('cfg:'):
        value = cfg.query(key[4:])
        if isinstance(value, tuple) and len(value) == 2 and value[0] is NOTSET:
            raise QLispError(f'Unknown config key: {key[4:]}')
        return value
    else:
        return key


def call_opaque(st: tuple, ctx: Context, lib: Library):
    name = gateName(st)
    gate, qubits = st
    type = None
    tmp_params = {}
    args = []

    if isinstance(gate, tuple):
        for arg in gate[1:]:
            if isinstance(arg, tuple) and isinstance(arg[0],
                                                     str) and arg[0] == 'with':
                for p in arg[1:]:
                    if isinstance(p, tuple) and len(p) == 2 and isinstance(
                            p[0], str):
                        if p[0] == 'type':
                            type = _try_to_lookup_config(ctx.cfg, p[1])
                        elif p[0].startswith('param:'):
                            tmp_params[p[0][6:]] = _try_to_lookup_config(
                                ctx.cfg, p[1])
            else:
                args.append(_try_to_lookup_config(ctx.cfg, arg))
    gatecfg = ctx.cfg._getGateConfig(name, *qubits, type=type)
    if gatecfg is None:
        gatecfg = GateConfig(name, qubits)

    tmp_params = {
        k: _try_to_call(v, gatecfg.params)
        for k, v in tmp_params.items()
    }
    params = gatecfg.params.copy()
    params.update(tmp_params)

    func, params_declaration = lib.getOpaque(name, gatecfg.type)
    if func is None:
        raise KeyError(f'Undefined {gatecfg.type} type of {name} opaque.')
    for p in params_declaration:
        if p.name not in params:
            pass
            # raise ValueError(
            #     f'{name} (type={gatecfg.type}) opaque of {qubits} missing parameter {k}.'
            # )

    args = tuple(args)

    sub_ctx = create_context(ctx, scopes=[*ctx.scopes, params])

    for cmd in func(sub_ctx, gatecfg.qubits, *args):
        _execute(sub_ctx, cmd)
    _ctx_update_biases(sub_ctx, ctx)
    _ctx_update_time(sub_ctx, ctx)
    _ctx_update_phases(sub_ctx, ctx)
    _ctx_update_waveforms(sub_ctx, ctx)
    _ctx_update_measurement_tasks(sub_ctx, ctx)


def _addWaveforms(ctx: Context, channel: tuple, wav: Waveform):
    name, *qubits = channel
    ch = ctx.cfg._getAWGChannel(name, *qubits)
    if isinstance(ch, AWGChannel):
        ctx.waveforms[ch.name] += wav
    else:
        _addMultChannelWaveforms(ctx, wav, ch)


def _addMultChannelWaveforms(ctx: Context, wav, ch: MultAWGChannel):
    lofreq = ch.lo_freq
    if ch.I is not None:
        try:
            I = (2 * wav * cos(-2 * pi * lofreq)).filter(high=2 * pi * lofreq)
        except:
            w = (2 * wav * cos(-2 * pi * lofreq))
            print("====== ERROR WAVEFORM ======")
            print("    lofreq =", lofreq)
            print("")
            print("    waveform =", w.tolist())
            print("====== ERROR WAVEFORM ======")
            raise
        ctx.waveforms[ch.I.name] += I
    if ch.Q is not None:
        Q = (2 * wav * sin(-2 * pi * lofreq)).filter(high=2 * pi * lofreq)
        ctx.waveforms[ch.Q.name] += Q


def _allocQubits(ctx, qlisp):
    for i, q in enumerate(ctx.cfg._getAllQubitLabels()):
        ctx.addressTable[q] = q
        ctx.addressTable[i] = q


def assembly_align_left(qlisp, ctx: Context, lib: Library):
    _allocQubits(ctx, qlisp)

    allQubits = set()

    for gate, qubits in qlisp:
        ctx.qlisp.append((gate, qubits))
        if isinstance(qubits, (int, str)):
            qubits = (ctx.qubit(qubits), )
        else:
            qubits = tuple(
                [ctx.qubit(q) if isinstance(q, int) else q for q in qubits])
        try:
            call_opaque((gate, qubits), ctx, lib=lib)
            allQubits.update(set(qubits))
        except:
            raise QLispError(f'assembly statement {(gate, qubits)} error.')
    call_opaque(('Barrier', tuple(allQubits)), ctx, lib=lib)
    for ch in ctx.biases:
        ctx.biases[ch] = 0
    _ctx_update_biases(ctx, ctx)
    ctx.end = max(ctx.time.values())

    code = QLispCode(cfg=ctx.cfg,
                     qlisp=ctx.qlisp,
                     waveforms=dict(ctx.waveforms),
                     measures=dict(ctx.measures),
                     end=ctx.end)
    return code


def assembly_align_right(qlisp, ctx: Context, lib: Library):
    raise NotImplementedError()

    _allocQubits(ctx, qlisp)

    allQubits = set()

    for gate, qubits in qlisp:
        ctx.qlisp.append((gate, qubits))
        if isinstance(qubits, (int, str)):
            qubits = (ctx.qubit(qubits), )
        else:
            qubits = tuple(
                [ctx.qubit(q) if isinstance(q, int) else q for q in qubits])
        try:
            call_opaque((gate, qubits), ctx, lib=lib)
            allQubits.update(set(qubits))
        except:
            raise QLispError(f'assembly statement {(gate, qubits)} error.')
    call_opaque(('Barrier', tuple(allQubits)), ctx, lib=lib)
    for ch in ctx.biases:
        ctx.biases[ch] = 0
    _ctx_update_biases(ctx, ctx)
    ctx.end = max(ctx.time.values())

    code = QLispCode(cfg=ctx.cfg,
                     qlisp=ctx.qlisp,
                     waveforms=dict(ctx.waveforms),
                     measures=dict(ctx.measures),
                     end=ctx.end)
    return code



import gzip
import hashlib
import itertools
import pickle
import time
import zipfile
from datetime import datetime
from functools import singledispatchmethod
from pathlib import Path
from typing import Any, Optional

import numpy as np
import pandas as pd
import xarray as xr
from sqlalchemy import (JSON, Boolean, Column, DateTime, Float, ForeignKey,
                        Integer, LargeBinary, Sequence, String, Table, Text,
                        create_engine)
from sqlalchemy.orm import (backref, declarative_base, relationship,
                            sessionmaker)
from sqlalchemy.orm.session import Session
from waveforms.security import InvalidKey, encryptPassword, verifyPassword

Base = declarative_base()

DATA_PATH = Path.home() / 'data'


def set_data_path(base_path: str) -> None:
    global DATA_PATH
    DATA_PATH = Path(base_path)


def get_data_path() -> Path:
    return DATA_PATH


# association table
user_roles = Table('user_roles', Base.metadata,
                   Column('user_id', ForeignKey('users.id'), primary_key=True),
                   Column('role_id', ForeignKey('roles.id'), primary_key=True))

record_reports = Table(
    'record_reports', Base.metadata,
    Column('record_id', ForeignKey('records.id'), primary_key=True),
    Column('report_id', ForeignKey('reports.id'), primary_key=True))

comment_tags = Table(
    'comment_tags', Base.metadata,
    Column('comment_id', ForeignKey('comments.id'), primary_key=True),
    Column('tag_id', ForeignKey('tags.id'), primary_key=True))

snapshot_tags = Table(
    'snapshot_tags', Base.metadata,
    Column('snapshot_id', ForeignKey('snapshots.id'), primary_key=True),
    Column('tag_id', ForeignKey('tags.id'), primary_key=True))

record_tags = Table(
    'record_tags', Base.metadata,
    Column('record_id', ForeignKey('records.id'), primary_key=True),
    Column('tag_id', ForeignKey('tags.id'), primary_key=True))

report_tags = Table(
    'report_tags', Base.metadata,
    Column('report_id', ForeignKey('reports.id'), primary_key=True),
    Column('tag_id', ForeignKey('tags.id'), primary_key=True))

sample_tags = Table(
    'sample_tags', Base.metadata,
    Column('sample_id', ForeignKey('samples.id'), primary_key=True),
    Column('tag_id', ForeignKey('tags.id'), primary_key=True))

sample_reports = Table(
    'sample_reports', Base.metadata,
    Column('sample_id', ForeignKey('samples.id'), primary_key=True),
    Column('report_id', ForeignKey('reports.id'), primary_key=True))

sample_records = Table(
    'sample_records', Base.metadata,
    Column('sample_id', ForeignKey('samples.id'), primary_key=True),
    Column('record_id', ForeignKey('records.id'), primary_key=True))

sample_comments = Table(
    'sample_comments', Base.metadata,
    Column('sample_id', ForeignKey('samples.id'), primary_key=True),
    Column('comment_id', ForeignKey('comments.id'), primary_key=True))

sample_transfer_comments = Table(
    'sample_transfer_comments', Base.metadata,
    Column('transfer_id', ForeignKey('sample_transfer.id'), primary_key=True),
    Column('comment_id', ForeignKey('comments.id'), primary_key=True))

report_comments = Table(
    'report_comments', Base.metadata,
    Column('report_id', ForeignKey('reports.id'), primary_key=True),
    Column('comment_id', ForeignKey('comments.id'), primary_key=True))

record_comments = Table(
    'record_comments', Base.metadata,
    Column('record_id', ForeignKey('records.id'), primary_key=True),
    Column('comment_id', ForeignKey('comments.id'), primary_key=True))


class Role(Base):
    __tablename__ = 'roles'

    id = Column(Integer, primary_key=True)
    name = Column(String)

    users = relationship('User', secondary=user_roles, back_populates='roles')

    def __repr__(self):
        return f"Role(name='{self.name}')"


class User(Base):
    __tablename__ = 'users'

    id = Column(Integer, primary_key=True)
    name = Column(String, unique=True)
    hashed_password = Column(LargeBinary(64))
    fullname = Column(String)

    roles = relationship('Role', secondary=user_roles, back_populates='users')
    attachments = relationship('Attachment', back_populates='user')
    comments = relationship('Comment', back_populates='user')

    def setPassword(self, password):
        self.hashed_password = encryptPassword(password)

    def verify(self, password):
        try:
            verifyPassword(password, self.hashed_password)
            return True
        except InvalidKey:
            return False

    def __repr__(self):
        return f"User(name='{self.name}')"


class ParameterEdge(Base):
    __tablename__ = 'parameter_edges'
    parent_id = Column(ForeignKey('parameters.id'), primary_key=True)
    child_id = Column(ForeignKey('parameters.id'), primary_key=True)
    key = Column(String(50), primary_key=True)


class Parameter(Base):
    __tablename__ = 'parameters'

    id = Column(Integer, primary_key=True)
    type = Column(String)
    unit = Column(String)
    integer = Column(Integer)
    real = Column(Float)
    imag = Column(Float)
    string = Column(String)
    buff = Column(LargeBinary)

    children = relationship("ParameterEdge",
                            foreign_keys=[ParameterEdge.parent_id],
                            backref="parent")
    parents = relationship("ParameterEdge",
                           foreign_keys=[ParameterEdge.child_id],
                           backref="child")

    @property
    def value(self):
        if self.type == 'integer':
            return self.integer
        elif self.type == 'real':
            return self.real
        elif self.type == 'complex':
            return self.real + 1j * self.imag
        elif self.type == 'string':
            return self.string
        elif self.type == 'buffer':
            return self.buff
        elif self.type == 'boolean':
            return self.integer == 1
        elif self.type is None or self.type == 'none':
            return None
        elif self.type == 'tree':
            return self.export()
        else:
            return pickle.loads(self.buff)

    @value.setter
    def value(self, value):
        self._set_value(value)

    @singledispatchmethod
    def _set_value(self, value):
        self.buff = pickle.dumps(value)
        self.type = 'pickle'

    @_set_value.register
    def _(self, value: int):
        self.integer = value
        self.type = 'integer'

    @_set_value.register
    def _(self, value: bool):
        self.integer = int(value)
        self.type = 'boolean'

    @_set_value.register
    def _(self, value: float):
        self.real = value
        self.type = 'real'

    @_set_value.register
    def _(self, value: complex):
        self.real = value.real
        self.imag = value.imag
        self.type = 'complex'

    @_set_value.register
    def _(self, value: str):
        self.string = value
        self.type = 'string'

    @_set_value.register
    def _(self, value: bytes):
        self.buff = value
        self.type = 'buffer'

    @_set_value.register
    def _(self, value: None):
        self.type = 'none'

    def export(self):
        if self.type == 'tree':
            return {a.key: a.child.export() for a in self.children}
        else:
            return self.value


class Snapshot(Base):
    __tablename__ = 'snapshots'

    id = Column(Integer, primary_key=True)
    root_id = Column(ForeignKey('parameters.id'))
    previous_id = Column(Integer, ForeignKey('snapshots.id'))

    followers = relationship("Snapshot",
                             backref=backref('previous', remote_side=[id]))

    root = relationship('Parameter')
    tags = relationship('Tag',
                        secondary=snapshot_tags,
                        back_populates='snapshots')

    def export(self):
        return self.root.export()


class ReportParameters(Base):
    __tablename__ = 'report_parameters'
    parent_id = Column(ForeignKey('reports.id'), primary_key=True)
    child_id = Column(ForeignKey('parameters.id'), primary_key=True)
    key = Column(String(50), primary_key=True)

    parameter = relationship('Parameter')

    @property
    def value(self):
        return self.parameter.value


class Tag(Base):
    __tablename__ = 'tags'

    id = Column(Integer, primary_key=True)
    text = Column(String, unique=True)

    comments = relationship('Comment',
                            secondary=comment_tags,
                            back_populates='tags')
    records = relationship('Record',
                           secondary=record_tags,
                           back_populates='tags')
    reports = relationship('Report',
                           secondary=report_tags,
                           back_populates='tags')
    samples = relationship('Sample',
                           secondary=sample_tags,
                           back_populates='tags')
    snapshots = relationship('Snapshot',
                             secondary=snapshot_tags,
                             back_populates='tags')

    def __init__(self, text) -> None:
        super().__init__()
        self.text = text

    def __repr__(self):
        return f"Tag('{self.text}')"


class Comment(Base):
    __tablename__ = 'comments'

    id = Column(Integer, primary_key=True)
    text = Column(String)
    user_id = Column(Integer, ForeignKey('users.id'))
    ctime = Column(DateTime, default=datetime.utcnow)
    mtime = Column(DateTime, default=datetime.utcnow)
    atime = Column(DateTime, default=datetime.utcnow)
    parent_id = Column(Integer, ForeignKey('comments.id'))

    replies = relationship("Comment", lazy="joined", join_depth=2)
    user = relationship('User', back_populates='comments')
    tags = relationship('Tag',
                        secondary=comment_tags,
                        back_populates='comments')
    attachments = relationship('Attachment', back_populates='comment')


class Attachment(Base):
    __tablename__ = 'attachments'

    id = Column(Integer, primary_key=True)
    filename = Column(String)
    mime_type = Column(String, default='application/octet-stream')
    user_id = Column(Integer, ForeignKey('users.id'))
    comment_id = Column(Integer, ForeignKey('comments.id'))
    ctime = Column(DateTime, default=datetime.utcnow)
    mtime = Column(DateTime, default=datetime.utcnow)
    atime = Column(DateTime, default=datetime.utcnow)
    size = Column(Integer)
    sha1 = Column(String)
    description = Column(Text)

    user = relationship('User', back_populates='attachments')
    comment = relationship('Comment', back_populates='attachments')

    @property
    def data(self) -> bytes:
        self.atime = datetime.utcnow()
        return _load_object(self.filename)

    @data.setter
    def data(self, data: bytes):
        self.mtime = datetime.utcnow()
        self.filename, self.sha1 = _save_object(data)


class InputText(Base):
    __tablename__ = 'inputs'

    id = Column(Integer, primary_key=True)
    hash = Column(LargeBinary(20))
    text_field = Column(Text, unique=True)

    @property
    def text(self):
        return self.text_field

    @text.setter
    def text(self, text):
        self.hash = hashlib.sha1(text.encode('utf-8')).digest()
        self.text_field = text

    def __repr__(self) -> str:
        return self.text


class Cell(Base):
    __tablename__ = 'cells'

    id = Column(Integer, primary_key=True)
    notebook_id = Column(Integer, ForeignKey("notebooks.id"))
    index = Column(Integer)
    ctime = Column(DateTime, default=datetime.utcnow)
    ftime = Column(DateTime, default=datetime.utcnow)
    input_id = Column(Integer, ForeignKey("inputs.id"))

    notebook = relationship("Notebook", back_populates="cells")
    input = relationship("InputText")

    def __repr__(self) -> str:
        return f"Cell(index={self.index}, input='{self.input}')"


class Notebook(Base):
    __tablename__ = 'notebooks'

    id = Column(Integer, primary_key=True)
    name = Column(String)
    ctime = Column(DateTime, default=datetime.utcnow)
    atime = Column(DateTime, default=datetime.utcnow)

    cells = relationship("Cell",
                         order_by=Cell.index,
                         back_populates="notebook")


class Sample(Base):
    __tablename__ = 'samples'

    id = Column(Integer, primary_key=True)
    name = Column(String)

    account_id = Column(Integer, ForeignKey("sample_accounts.id"))

    tags = relationship("Tag", secondary=sample_tags, back_populates="samples")
    records = relationship("Record",
                           secondary=sample_records,
                           back_populates="samples")
    reports = relationship("Report",
                           secondary=sample_reports,
                           back_populates="samples")
    transfers = relationship("SampleTransfer",
                             order_by="SampleTransfer.ctime",
                             back_populates="sample")
    account = relationship("SampleAccount", back_populates="samples")
    comments = relationship("Comment", secondary=sample_comments)


class SampleAccountType(Base):
    __tablename__ = 'sample_account_types'
    id = Column(Integer, primary_key=True)
    name = Column(String)
    description = Column(String)

    accounts = relationship("SampleAccount", back_populates="type")


class SampleAccount(Base):
    __tablename__ = 'sample_accounts'

    id = Column(Integer, primary_key=True)
    name = Column(String)
    type_id = Column(Integer, ForeignKey("sample_account_types.id"))
    description = Column(String)

    type = relationship("SampleAccountType", back_populates="accounts")

    samples = relationship("Sample", back_populates="account")


class SampleTransfer(Base):
    __tablename__ = 'sample_transfer'

    id = Column(Integer, primary_key=True)
    sample_id = Column(Integer, ForeignKey("samples.id"))
    user_id = Column(Integer, ForeignKey("users.id"))
    ctime = Column(DateTime, default=datetime.utcnow)
    debtor_id = Column(Integer, ForeignKey("sample_accounts.id"))
    creditor_id = Column(Integer, ForeignKey("sample_accounts.id"))

    user = relationship("User")
    sample = relationship("Sample", back_populates="transfers")
    debtor = relationship("SampleAccount", foreign_keys=[debtor_id])
    creditor = relationship("SampleAccount", foreign_keys=[creditor_id])
    comments = relationship("Comment", secondary=sample_transfer_comments)


class Record(Base):
    __tablename__ = 'records'

    id = Column(Integer, primary_key=True)
    ctime = Column(DateTime, default=datetime.utcnow)
    mtime = Column(DateTime, default=datetime.utcnow)
    atime = Column(DateTime, default=datetime.utcnow)
    user_id = Column(Integer, ForeignKey('users.id'))
    parent_id = Column(Integer, ForeignKey('records.id'))
    cell_id = Column(Integer, ForeignKey('cells.id'))

    app = Column(String)
    file = Column(String)
    key = Column(String)
    config = Column(JSON)
    task_hash = Column(LargeBinary(32))

    parent = relationship("Record",
                          remote_side=[id],
                          back_populates="children")
    children = relationship("Record",
                            remote_side=[parent_id],
                            back_populates="parent")

    user = relationship("User")
    samples = relationship("Sample",
                           secondary=sample_records,
                           back_populates="records")
    cell = relationship("Cell")

    reports = relationship('Report',
                           secondary=record_reports,
                           back_populates='records')
    tags = relationship('Tag', secondary=record_tags, back_populates='records')
    comments = relationship('Comment', secondary=record_comments)

    def __init__(self,
                 file: str = None,
                 key: Optional[str] = None,
                 dims: list[str] = [],
                 vars: list[str] = [],
                 dims_units: list[str] = [],
                 vars_units: list[str] = [],
                 coords: Optional[dict] = None):
        self.file = file
        if key is None:
            self.key = '/Data' + time.strftime("%Y%m%d%H%M%S")
        else:
            self.key = key
        self.ds: xr.Dataset = None
        self.df: pd.DataFrame = None
        self.dims = dims
        self.vars = vars
        self.dims_units = dims_units
        self.vars_units = vars_units
        self.coords = coords

        self._buff = ([], [])

    @property
    def data(self):
        self.atime = datetime.utcnow()
        ret = _load_object(self.file)
        if isinstance(ret, dict) and 'meta' in ret:
            ret['meta']['id'] = self.id
        return ret

    @data.setter
    def data(self, data):
        self.file, _ = _save_object(data)


class Report(Base):
    __tablename__ = 'reports'

    id = Column(Integer, primary_key=True)
    ctime = Column(DateTime, default=datetime.utcnow)
    mtime = Column(DateTime, default=datetime.utcnow)
    atime = Column(DateTime, default=datetime.utcnow)
    user_id = Column(Integer, ForeignKey('users.id'))

    category = Column(String)
    title = Column(String)
    content = Column(Text)
    file = Column(String)
    key = Column(String)
    task_hash = Column(LargeBinary(32))

    user = relationship("User")
    samples = relationship("Sample",
                           secondary=sample_reports,
                           back_populates="reports")

    records = relationship('Record',
                           secondary=record_reports,
                           back_populates='reports')

    tags = relationship('Tag', secondary=report_tags, back_populates='reports')
    comments = relationship('Comment', secondary=report_comments)
    _parameters = relationship('ReportParameters')

    @property
    def obj(self):
        self.atime = datetime.utcnow()
        return _load_object(self.file)

    @obj.setter
    def obj(self, data):
        self.file, _ = _save_object(data)

    def parameters(self):
        return {p.key: p.value for p in self._parameters}

    def set_parameter(self, **kwds):
        for key, value in kwds.items():
            for p in self._parameters:
                if p.key == key:
                    p.parameter.value = value
                    break
            else:
                p = Parameter()
                p.value = value
                rp = ReportParameters(key=key)
                rp.parameter = p
                self._parameters.append(rp)


def create_tables(engine, tables_only=False):
    Base.metadata.create_all(engine)
    if tables_only:
        return

    sys_role = Role(name='sys')
    kernel = User(name='BIG BROTHER')
    kernel.roles.append(sys_role)

    root_role = Role(name='root')
    admin_role = Role(name='admin')
    root_user = User(name='root')
    root_user.setPassword('123')
    root_user.roles.append(root_role)
    root_user.roles.append(admin_role)

    guest_role = Role(name='guest')
    guest_user = User(name='guest')
    guest_user.setPassword('')
    guest_user.roles.append(guest_role)

    t1 = SampleAccountType(name='factory')
    t2 = SampleAccountType(name='destroyed')
    t3 = SampleAccountType(name='storage')
    t4 = SampleAccountType(name='fridge')
    a = SampleAccount(name='destroyed')
    a.type = t2

    Session = sessionmaker(bind=engine)
    session = Session()

    session.add(root_user)
    session.add(guest_user)
    session.add(kernel)
    session.add_all([t1, t2, t3, t4, a])
    try:
        session.commit()
    except:
        session.rollback()


def _save_object(data: Any) -> tuple[str, str]:
    data = pickle.dumps(data)
    hashstr = hashlib.sha1(data).hexdigest()
    file = get_data_path(
    ) / 'objects' / hashstr[:2] / hashstr[2:4] / hashstr[4:]
    file.parent.mkdir(parents=True, exist_ok=True)
    with gzip.open(file, 'wb') as f:
        f.write(data)
    return str('/'.join(file.parts[-4:])), hashstr


def _load_object(file: str) -> bytes:
    with gzip.open(get_data_path() / file, 'rb') as f:
        data = pickle.load(f)
    return data



import pytest
from waveforms.qlisp import (Library, Parameter, QLispCode, compile, libraries,
                             stdlib)

from config import config

qasm = """
OPENQASM 2.0;

include "qelib1.inc";
opaque iSWAP c,t;

gate createBellPair a, b
{
   h a;
   cx a,b;
}

gate bellMeasure a, b
{
   cx a, b;
   h a;
}

qreg q[2];
creg c[2];

createBellPair q[0], q[1];
iSWAP q[0],q[1];
bellMeasure q[0],q[1];
barrier q;
measure q -> c;
"""

qlisp = [
    ('createBellPair', (0, 1)),
    ('iSWAP', (0, 1)),
    ('bellMeasure', (0, 1)),
    ('Barrier', (0, 1)),
    (('Measure', 0), 0),
    (('Measure', 1), 1),
]

qlisp3 = [
    ('createBellPair', (0, 1)),
    (('iSWAP', ('with', ('param:amp', 0.5))), (0, 1)),
    ('bellMeasure', (0, 1)),
    ('Barrier', (0, 1)),
    (('Measure', 0), 0),
    (('Measure', 1), 1),
]

qlisp4 = [
    ('define', 'Q1', 0),
    ('define', 'Q2', 1),
    ('createBellPair', ('Q1', 'Q2')),
    (('iSWAP', ('with', ('param:amp', lambda amp: amp - 0.3))), ('Q1', 'Q2')),
    ('bellMeasure', (0, 1)),
    ('Barrier', (0, 1)),
    (('Measure', 0), 0),
    (('Measure', 1), 1),
]


@pytest.fixture
def cfg():
    from waveforms.qlisp.config import Config
    try:
        yield Config.fromdict(config)
    finally:
        pass


@pytest.fixture
def lib():
    from numpy import pi

    lib = libraries(stdlib)

    @lib.opaque('CZ',
                params=[
                    Parameter('duration', float, 50e-9, 's'),
                    Parameter('amp', float, 0.8, 'a.u.')
                ])
    def CZ(ctx, qubits):
        control, target = qubits
        from waveforms.waveform import cos, square

        duration = ctx.params['duration']
        amp = ctx.params['amp']
        t = max(ctx.time[control], ctx.time[target])

        if amp > 0 and duration > 0:
            pulse = (cos(pi / duration) * square(duration)) >> duration / 2
            yield ('!add', 'waveform', amp * pulse >> t), ('coupler.Z',
                                                           control, target)
        yield ('!add', 'time', t + duration), control
        yield ('!add', 'time', t + duration), target

    @lib.opaque('iSWAP',
                params=[('duration', float, 50e-9, 's'),
                        ('amp', float, 0.8, 'a.u.'),
                        ('offset', float, 0, 'a.u.'),
                        ('frequency', float, 10e6, 'Hz')])
    def iSWAP(ctx, qubits):
        from waveforms.waveform import sin, square

        control, target = qubits

        duration = ctx.params['duration']
        amp = ctx.params['amp']
        offset = ctx.params['offset']
        frequency = ctx.params['frequency']

        t = max(ctx.time[control], ctx.time[target])
        if duration > 0:
            pulse = square(duration) >> duration / 2
            pulse = pulse * offset + amp * pulse * sin(2 * pi * frequency)
            yield ('!add', 'waveform', amp * pulse >> t), ('coupler.Z',
                                                           control, target)
        yield ('!add', 'time', t + duration), control
        yield ('!add', 'time', t + duration), target

    @lib.gate(2)
    def bellMeasure(qubits):
        a, b = qubits
        yield ('Cnot', (a, b))
        yield ('H', a)

    @lib.gate(2)
    def createBellPair(qubits):
        a, b = qubits
        yield ('H', a)
        yield ('Cnot', (a, b))

    try:
        yield lib
    finally:
        pass


def test_lib(lib):
    assert isinstance(lib, Library)


def test_compile(lib, cfg):
    ret = compile(qasm, cfg=cfg, lib=lib)
    assert isinstance(ret, QLispCode)
    ret2 = compile(qlisp, cfg=cfg, lib=lib)
    assert isinstance(ret2, QLispCode)
    ret3 = compile(qlisp3, cfg=cfg, lib=lib)
    assert isinstance(ret3, QLispCode)
    ret4 = compile(qlisp4, cfg=cfg, lib=lib)
    assert isinstance(ret4, QLispCode)
    assert 'AWG.Z' in ret.waveforms
    for k, wav in ret.waveforms.items():
        assert k in ret2.waveforms
        assert wav == ret2.waveforms[k]
        if k == 'AWG.Z':
            assert wav != ret3.waveforms[k]
            assert wav != ret4.waveforms[k]
            assert ret3.waveforms[k] == ret4.waveforms[k]
        else:
            assert wav == ret3.waveforms[k]
            assert wav == ret4.waveforms[k]



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Node for an OPENQASM external function."""

import warnings
import numpy as np

from .node import Node
from .nodeexception import NodeException


class External(Node):
    """Node for an OPENQASM external function.
    children[0] is an id node with the name of the function.
    children[1] is an expression node.
    """

    def __init__(self, children):
        """Create the external node."""
        super().__init__('external', children, None)

    def qasm(self, prec=None):
        """Return the corresponding OPENQASM string."""
        if prec is not None:
            warnings.warn('Parameter \'External.qasm(..., prec)\' is no longer used and is being '
                          'deprecated.', DeprecationWarning, 2)
        return self.children[0].qasm() + "(" + self.children[1].qasm() + ")"

    def latex(self, prec=None, nested_scope=None):
        """Return the corresponding math mode latex string."""
        if prec is not None:
            warnings.warn('Parameter \'External.latex(..., prec)\' is no longer used and is being '
                          'deprecated.', DeprecationWarning, 2)
        if nested_scope is not None:
            warnings.warn('Parameter \'External.latex(..., nested_scope)\' is no longer used and '
                          'is being deprecated.', DeprecationWarning, 2)
        try:
            from pylatexenc.latexencode import utf8tolatex
        except ImportError:
            raise ImportError("To export latex from qasm "
                              "pylatexenc needs to be installed. Run "
                              "'pip install pylatexenc' before using this "
                              "method.")
        return utf8tolatex(self.sym())

    def real(self, nested_scope=None):
        """Return the correspond floating point number."""
        op = self.children[0].name
        expr = self.children[1]
        dispatch = {
            'sin': np.sin,
            'cos': np.cos,
            'tan': np.tan,
            'asin': np.arcsin,
            'acos': np.arccos,
            'atan': np.arctan,
            'exp': np.exp,
            'ln': np.log,
            'sqrt': np.sqrt
        }
        if op in dispatch:
            arg = expr.real(nested_scope)
            return dispatch[op](arg)
        else:
            raise NodeException("internal error: undefined external")

    def sym(self, nested_scope=None):
        """Return the corresponding symbolic expression."""
        op = self.children[0].name
        expr = self.children[1]
        dispatch = {
            'sin': np.sin,
            'cos': np.cos,
            'tan': np.tan,
            'asin': np.arcsin,
            'acos': np.arccos,
            'atan': np.arctan,
            'exp': np.exp,
            'ln': np.log,
            'sqrt': np.sqrt
        }
        if op in dispatch:
            arg = expr.sym(nested_scope)
            return dispatch[op](arg)
        else:
            raise NodeException("internal error: undefined external")


from typing import Literal, Optional, Sequence, Union

from .arch import get_arch
from .assembly import assembly_align_left, assembly_align_right, call_opaque
from .config import Config
from .library import Library, libraries
from .macro import extend_macro, reduceVirtualZ
from .qasm import qasm_eval
from .base import Context, create_context, getConfig
from .libs import std


def assembly(qlisp,
             cfg: Optional[Config] = None,
             lib: Library = std,
             ctx: Optional[Context] = None,
             align: Literal['left', 'right'] = 'left'):

    if cfg is None:
        cfg = getConfig()

    if ctx is None:
        ctx = create_context(cfg=cfg)

    if align == 'left':
        return assembly_align_left(qlisp, ctx, lib)
    elif align == 'right':
        return assembly_align_right(qlisp, ctx, lib)
    else:
        raise ValueError(f'align={align} is not supported.')


def compile(prog,
            cfg: Optional[Config] = None,
            lib: Union[Library, Sequence[Library]] = std,
            **options):
    """
    options: 
        qasm_only = True: only compile qasm to qlisp
        no_virtual_z = True: keep P gates as original form.
        no_assembly = True: return simplified qlisp.
    """

    if isinstance(lib, Library):
        lib = lib
    else:
        lib = libraries(*lib)

    if isinstance(prog, str):
        prog = qasm_eval(prog, lib)
    if 'qasm_only' in options:
        return list(prog)
    prog = extend_macro(prog, lib)
    if 'no_virtual_z' in options:
        return list(prog)
    prog = reduceVirtualZ(prog, lib)
    if 'no_link' in options:
        return list(prog)
    if 'align_right' in options or 'ar' in options:
        align = 'right'
    else:
        align = 'left'
    code = assembly(prog, cfg, lib, align=align)
    if 'arch' in options:
        code.arch = options['arch']
    if 'shots' in options:
        code.shots = options['shots']
    if 'signal' in options:
        code.signal = options['signal']
    if code.arch == 'general' or 'no_assembly' in options:
        return code
    return get_arch(code.arch).assembly_code(code)



from typing import Hashable, Optional

import numpy as np
from scipy.optimize import minimize
from scipy.signal import correlate
from scipy.sparse import coo_matrix, diags, linalg


def fit_relative_delay(waveform, data, sample_rate, fit=True):
    t = np.linspace(0, len(data) / sample_rate, len(data), endpoint=False)
    if isinstance(waveform, np.ndarray):
        fit = False
        corr = correlate(waveform, data, mode='same', method='fft')
    else:
        corr = correlate(waveform(t), data, mode='same', method='fft')
    delay = 0.5 * (t[0] + t[-1]) - t[np.argmax(corr)]

    if not fit:
        return delay

    def fun(delay, t, waveform, sig):
        if isinstance(delay, (int, float, complex)):
            ret = -correlate((waveform >> delay)(t), sig, mode='valid')[0]
            return ret
        else:
            return np.array([fun(d, t, waveform, sig) for d in delay])

    ret = minimize(fun, [delay], args=(t, waveform, data))
    return ret.x[0]


def calc_delays(relative_delays: dict[tuple[str, str], float],
                reference: float = 0,
                reference_channel: Optional[str] = None,
                full: bool = False) -> dict[str, float]:
    channels = []
    channels_map = {}

    def new_channel(ch):
        if ch not in channels:
            channels_map[ch] = len(channels)
            channels.append(ch)

    matrix = []
    weight = []
    absolute_error = True
    y = []
    for i, ((ch1, ch2), delay) in enumerate(relative_delays.items()):
        if ch1 not in channels:
            new_channel(ch1)
        if ch2 not in channels:
            new_channel(ch2)
        matrix.append([i, channels_map[ch1], 1])
        matrix.append([i, channels_map[ch2], -1])
        if isinstance(delay, (int, float)):
            y.append(delay)
            weight.append(1)
            absolute_error = False
        else:
            y.append(delay[0])
            weight.append(1 / delay[1]**2)

    if reference_channel is None:
        reference_channel = channels[0]
    new_channel(reference_channel)
    matrix.append([i + 1, channels_map[reference_channel], 1])
    y.append(0)
    weight.append(np.max(weight))

    row, col, data = list(zip(*matrix))

    X = coo_matrix((data, (row, col)), shape=(len(y), len(channels)))
    W = diags(weight, 0, shape=(len(y), len(y)))

    M = linalg.inv(X.T @ W @ X)

    beta = M @ X.T @ W @ y
    offset = reference - beta[channels_map[reference_channel]]

    if full:
        if absolute_error:
            pass
        else:
            r = X @ beta - y
            chi_square = r.T @ W @ r

    return {ch: v + offset for ch, v in zip(channels, beta)}



from waveforms.quantum.clifford import cliffordOrder


def test_cliffordOrder():
    assert cliffordOrder(0) == 1
    assert cliffordOrder(1) == 24
    assert cliffordOrder(2) == 11520
    assert cliffordOrder(3) == 92897280



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Node for an OPENQASM integer."""

import warnings

from .node import Node


class Int(Node):
    """Node for an OPENQASM integer.
    This node has no children. The data is in the value field.
    """

    def __init__(self, id):
        """Create the integer node."""
        # pylint: disable=redefined-builtin
        super().__init__("int", None, None)
        self.value = id

    def to_string(self, indent):
        """Print with indent."""
        ind = indent * ' '
        print(ind, 'int', self.value)

    def qasm(self, prec=None):
        """Return the corresponding OPENQASM string."""
        if prec is not None:
            warnings.warn('Parameter \'Int.qasm(..., prec)\' is no longer used and is being '
                          'deprecated.', DeprecationWarning, 2)
        return "%d" % self.value

    def latex(self, prec=None, nested_scope=None):
        """Return the corresponding math mode latex string."""
        if prec is not None:
            warnings.warn('Parameter \'Int.latex(..., prec)\' is no longer used and is being '
                          'deprecated.', DeprecationWarning, 2)
        if nested_scope is not None:
            warnings.warn('Parameter \'Int.latex(..., nested_scope)\' is no longer used and is '
                          'being deprecated.', DeprecationWarning, 2)
        return "%d" % self.value

    def sym(self, nested_scope=None):
        """Return the correspond symbolic number."""
        del nested_scope
        return float(self.value)

    def real(self, nested_scope=None):
        """Return the correspond floating point number."""
        del nested_scope  # ignored
        return float(self.value)


from typing import Any


class COMMAND():
    """Commands for the executor"""
    __slots__ = ('address', 'value')

    def __init__(self, address: str, value: Any):
        self.address = address
        self.value = value


class READ(COMMAND):
    """Read a value from the scheduler"""

    def __init__(self, address: str):
        super().__init__(address, 'READ')

    def __repr__(self) -> str:
        return f"READ({self.address})"


class WRITE(COMMAND):

    def __repr__(self) -> str:
        return f"WRITE({self.address}, {self.value})"


class TRIG(COMMAND):
    """Trigger the system"""

    def __init__(self, address: str):
        super().__init__(address, 0)

    def __repr__(self) -> str:
        return f"TRIG({self.address})"


class SYNC(COMMAND):
    """Synchronization command"""

    def __init__(self, delay: float = 0):
        super().__init__('SYNC', delay)

    def __repr__(self) -> str:
        return f"SYNC({self.value})"


class PUSH(COMMAND):

    def __init__(self):
        super().__init__('PUSH', 0)

    def __repr__(self) -> str:
        return f"PUSH()"


class FREE(COMMAND):

    def __init__(self):
        super().__init__('FREE', 0)

    def __repr__(self) -> str:
        return f"FREE()"


CommandList = list[COMMAND]
DataMap = dict[str, dict]
RawData = Any
Result = dict



from itertools import repeat

import numpy as np


def DD(qubit, t, gates, pos, f=0):
    seq = [('X/2', qubit)]
    i = 0
    for gate in gates:
        gap = t * (pos[i] - pos[i - 1]) if i > 0 else t * pos[0]
        seq.append((('Delay', gap), qubit))
        seq.append((gate, qubit))
        i += 1
    gap = t * (1 - pos[-1]) if len(pos) > 0 else t
    seq.append((('Delay', gap), qubit))
    if f != 0:
        seq.append((('P', 2 * np.pi * f * t), qubit))
    seq.append(('X/2', qubit))
    return seq


def XY4(qubit, t, f=0):
    pos = np.arange(1, 5) / 5
    return DD(qubit, t, ['X', 'Y', 'X', 'Y'], pos, f)


def XY8(qubit, t, f=0):
    pos = np.arange(1, 9) / 9
    return DD(qubit, t, ['X', 'Y', 'X', 'Y', 'Y', 'X', 'Y', 'X'], pos, f)


def XY16(qubit, t, f=0):
    pos = np.arange(1, 17) / 17
    return DD(qubit, t, [
        'X', 'Y', 'X', 'Y', 'Y', 'X', 'Y', 'X', 'X', 'Y', 'X', 'Y', 'Y', 'X',
        'Y', 'X'
    ], pos, f)


def UDD(qubit, n, t, f=0):
    j = np.arange(n) + 1
    return DD(qubit, t, repeat('Y', times=n),
              np.sin(np.pi * j / (2 * n + 2))**2, f)


def CPMG(qubit, n, t, f=0):
    j = np.arange(n) + 1
    return DD(qubit, t, repeat('Y', times=n), (j - 0.5) / n, f)


def CP(qubit, n, t, f=0):
    j = np.arange(n) + 1
    return DD(qubit, t, repeat('X', times=n), (j - 0.5) / n, f)


def Ramsey(qubit, t, f=0):
    return [('X/2', qubit), (('Delay', t), qubit),
            (('rfUnitary', np.pi / 2, 2 * np.pi * f * t), qubit)]


def SpinEcho(qubit, t, f=0):
    return [('X/2', qubit), (('Delay', t / 2), qubit),
            (('rfUnitary', np.pi, np.pi * f * t), qubit),
            (('Delay', t / 2), qubit), ('X/2', qubit)]


_ALLXYSeq = [('I', 'I'), ('X', 'X'), ('Y', 'Y'), ('X', 'Y'), ('Y', 'X'),
             ('X/2', 'I'), ('Y/2', 'I'), ('X/2', 'Y/2'), ('Y/2', 'X/2'),
             ('X/2', 'Y'), ('Y/2', 'X'), ('X', 'Y/2'), ('Y', 'X/2'),
             ('X/2', 'X'), ('X', 'X/2'), ('Y/2', 'Y'), ('Y', 'Y/2'),
             ('X', 'I'), ('Y', 'I'), ('X/2', 'X/2'), ('Y/2', 'Y/2')]


def ALLXY(qubit, i):
    assert 0 <= i < len(
        _ALLXYSeq), f"i={i} is out of range(0, {len(_ALLXYSeq)})"
    return [(gate, qubit) for gate in _ALLXYSeq[i]]



from itertools import repeat
from typing import Optional, Sequence

import numpy as np


def getFTMatrix(fList: Sequence[float],
                numOfPoints: int,
                phaseList: Optional[Sequence[float]] = None,
                weight: Optional[np.ndarray] = None,
                sampleRate: float = 1e9) -> np.ndarray:
    """
    get a matrix for Fourier transform

    Args:
        fList (Sequence[float]): list of frequencies
        numOfPoints (int): size of signal frame
        phaseList (Optional[Sequence[float]], optional): list of phase. Defaults to None.
        weight (Optional[np.ndarray], optional): weight or list of weight. Defaults to None.
        sampleRate (float, optional): sample rate of signal. Defaults to 1e9.

    Returns:
        numpy.ndarray: exp matrix
    
    >>> shots, numOfPoints, sampleRate = 100, 1000, 1e9
    >>> f1, f2 = -12.7e6, 32.8e6
    >>> signal = np.random.randn(shots, numOfPoints)
    >>> e = getFTMatrix([f1, f2], numOfPoints, sampleRate=sampleRate)
    >>> ret = signal @ e
    >>> ret.shape
    (100, 2)
    >>> t = np.arange(numOfPoints) / sampleRate
    >>> signal = 0.8 * np.sin(2 * np.pi * f1 * t) + 0.2 * np.cos(2 * np.pi * f2 * t)
    >>> signal @ e
    array([-0.00766509-0.79518987j,  0.19531432+0.00207068j])
    >>> spec = 2 * np.fft.fft(signal) / numOfPoints
    >>> freq = np.fft.fftfreq(numOfPoints)
    >>> e = getFTMatrix(freq, numOfPoints, sampleRate=1)
    >>> np.allclose(spec, signal @ e)
    True
    """
    e = []
    t = np.linspace(0, numOfPoints / sampleRate, numOfPoints, endpoint=False)
    if weight is None or len(weight) == 0:
        weight = np.full(numOfPoints, 2 / numOfPoints)
    if phaseList is None or len(phaseList) == 0:
        phaseList = np.zeros_like(fList)
    if weight.ndim == 1:
        weightList = repeat(weight)
    else:
        weightList = weight
    for f, phase, weight in zip(fList, phaseList, weightList):
        e.append(weight * np.exp(-1j * (2 * np.pi * f * t + phase)))
    return np.asarray(e).T


if __name__ == "__main__":
    import doctest
    doctest.testmod()



from typing import Hashable


def graphs(
    edges: list[tuple[Hashable, Hashable]]
) -> list[list[tuple[Hashable, Hashable]]]:
    """
    graphs

    Args:
        edges: list of edges

    Returns:
        list of graphs
    """
    groups = []
    ret = []

    for a, b in edges:

        group_a_index, group_b_index = None, None
        for i, group in enumerate(groups):
            if a in group and b in group:
                break
            elif a in group:
                group_a_index = i
            elif b in group:
                group_b_index = i
            else:
                continue
            if group_a_index is not None and group_b_index is not None:
                group_a_index, group_b_index = sorted(
                    [group_a_index, group_b_index])
                group_b = groups.pop(group_b_index)
                group_a = groups.pop(group_a_index)
                groups.append(group_a | group_b)
                x = ret.pop(group_b_index)
                y = ret.pop(group_a_index)
                ret.append([*x, *y, (a, b)])
                break
        else:
            if group_a_index is None and group_b_index is None:
                groups.append({a, b})
                ret.append([(a, b)])
            elif group_a_index is None:
                groups[group_b_index].add(a)
                ret.append([(a, b)])
            elif group_b_index is None:
                groups[group_a_index].add(b)
                ret.append([(a, b)])

    return ret


def minimum_spanning_tree(
    edges: list[tuple[Hashable, Hashable]]
    | list[tuple[Hashable, Hashable, float]]
) -> list[tuple[Hashable, Hashable]]:
    """
    minimum spanning tree

    Args:
        edges: list of edges

    Returns:
        list of edges in minimum spanning tree
    """
    if not edges:
        return []

    if isinstance(edges[0], tuple) and len(edges[0]) == 2:
        edges = [(u, v, 1) for u, v in edges]

    nodes = set()
    for u, v, _ in edges:
        nodes.add(u)
        nodes.add(v)

    nodes = list(nodes)
    nodes_map = {n: i for i, n in enumerate(nodes)}

    edges = [(nodes_map[u], nodes_map[v], w) for u, v, w in edges]
    edges.sort(key=lambda e: e[2])

    tree = []
    parent = list(range(len(nodes)))
    rank = [0] * len(nodes)

    def find(u):
        if parent[u] != u:
            parent[u] = find(parent[u])
        return parent[u]

    def union(u, v):
        u = find(u)
        v = find(v)
        if u == v:
            return
        if rank[u] < rank[v]:
            u, v = v, u
        parent[v] = u
        if rank[u] == rank[v]:
            rank[u] += 1

    for u, v, w in edges:
        if find(u) != find(v):
            union(u, v)
            tree.append((nodes[u], nodes[v]))

    return tree



from typing import Type, Union
from sqlalchemy.orm import Session
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.orm import aliased, Query
from .models import (Cell, Comment, InputText, InvalidKey, Notebook, Record,
                     Report, Role, Sample, SampleAccount, SampleTransfer, Tag,
                     User)


def create_user(session: Session, user_name: str, password: str) -> User:
    """Create a user in the database."""
    user = User(name=user_name)
    user.setPassword(password)
    session.add(user)
    return user


def login(session: Session, user_name: str, password: str) -> User:
    """Login a user in the database."""
    try:
        user = session.query(User).filter(User.name == user_name).one()
    except NoResultFound:
        return None
    try:
        user.verify(password)
    except InvalidKey:
        return None
    return user


def get_all_records(session: Session) -> list:
    """Get all records from the database."""
    return session.query(Record).all()


def get_all_samples(session: Session) -> list:
    """Get all samples from the database."""
    return session.query(Sample).all()


def get_all_tags(session: Session) -> list:
    """Get all tags from the database."""
    return session.query(Tag).all()


def get_all_users(session: Session) -> list:
    """Get all users from the database."""
    return session.query(User).all()


def get_all_roles(session: Session) -> list:
    """Get all roles from the database."""
    return session.query(Role).all()


def get_all_notebooks(session: Session) -> list:
    """Get all notebooks from the database."""
    return session.query(Notebook).all()


def get_all_reports(session: Session) -> list:
    """Get all reports from the database."""
    return session.query(Report).all()


def tag(session: Session, tag_text: str) -> Tag:
    """Get a tag from the database or create a new if not exists."""
    try:
        return session.query(Tag).filter(Tag.text == tag_text).one()
    except NoResultFound:
        tag = Tag(text=tag_text)
        return tag


def tag_it(session: Session, tag_text: str, obj: Union[Sample, Record,
                                                      Report]) -> Tag:
    """Tag an object."""
    if obj.id is None:
        session.add(obj)
        obj.tags.append(tag(session, tag_text))
    else:
        session.query(type(obj)).filter(
            type(obj).id == obj.id).one().tags.append(tag(session, tag_text))
    session.commit()


def get_object_with_tags(session: Session,
                         cls: Union[Type[Comment], Type[Sample], Type[Record],
                                    Type[Report]], *tags: str) -> Query:
    """
    Query objects with the given tags.

    Parameters
    ----------
    session : :class:`sqlalchemy.orm.Session`
        The database session.
    cls : :class:`sqlalchemy.orm.Mapper`
        The object class.
    tags : str
        The tags.

    Returns
    -------
    :class:`sqlalchemy.orm.Query`
        The query.
    """
    if isinstance(session, Query):
        q = session
    else:
        q = session.query(cls)
    if not hasattr(cls, 'tags'):
        return []

    aliase = {tag: aliased(Tag) for tag in tags}

    for tag, a in aliase.items():
        q = q.join(a, cls.tags)
        q = q.filter(a.text == tag)
    return q


def create_notebook(session: Session, notebook_name: str) -> Notebook:
    """Create a notebook in the database."""
    notebook = Notebook(name=notebook_name)
    session.add(notebook)
    return notebook


def create_sample(session: Session, sample_name: str, notebook_id: int,
                  record_id: int) -> Sample:
    """Create a sample in the database."""
    sample = Sample(name=sample_name,
                    notebook_id=notebook_id,
                    record_id=record_id)
    session.add(sample)
    return sample


def create_record(session: Session, record_name: str, notebook_id: int,
                  sample_id: int) -> Record:
    """Create a record in the database."""
    record = Record(name=record_name,
                    notebook_id=notebook_id,
                    sample_id=sample_id)
    session.add(record)
    return record


def create_role(session: Session, role_name: str) -> Role:
    """Create a role in the database."""
    role = Role(name=role_name)
    session.add(role)
    return role


def create_cell(session: Session, notebook: Notebook, input_text: str) -> Cell:
    """Create a cell in the database."""
    cell = Cell()
    cell.notebook = notebook
    cell.input = create_input_text(session, input_text)
    cell.index = len(notebook.cells) - 1
    session.add(cell)
    notebook.atime = cell.ctime
    return cell


def create_input_text(session: Session, input_text: str) -> InputText:
    """Create an input text in the database."""
    input = InputText()
    input.text = input_text
    try:
        input = session.query(InputText).filter(
            InputText.hash == input.hash,
            InputText.text_field == input_text).one()
    except NoResultFound:
        session.add(input)
    return input


def create_sample_account(session: Session,
                          account_name: str) -> SampleAccount:
    """Create a sample account in the database."""
    account = SampleAccount(name=account_name)
    session.add(account)
    return account


def transform_sample(session: Session, sample: Sample, source: SampleAccount,
                     dest: SampleAccount) -> None:
    """Transform a sample from one account to another."""
    if (sample.account_id is None and
            source.type.name != 'factory') or sample.account_id != source.id:
        raise ValueError('Sample is not in the source account.')
    try:
        sample.account = dest
        transfer = SampleTransfer()
        transfer.sample = sample
        transfer.debtor = source
        transfer.creditor = dest
        session.add(transfer)
        session.commit()
        return transfer
    except:
        session.rollback()
        raise



from pathlib import Path

from numpy import mod, pi
from waveforms.waveform import (cos, cosPulse, gaussian, mixing, sin, square,
                                zero)
from waveforms.waveform_parser import wave_eval

from ..library import Library, Parameter
from ..base import MeasurementTask

EPS = 1e-9

std = Library()
std.qasmLib = {
    'qelib1.inc': Path(__file__).parent.parent / 'qasm' / 'libs' / 'qelib1.inc'
}


@std.gate()
def u3(qubit, theta, phi, lambda_):
    yield (('U', theta, phi, lambda_), qubit)


@std.gate()
def u2(qubit, phi, lambda_):
    yield (('U', pi / 2, phi, lambda_), qubit)


@std.gate()
def u1(qubit, lambda_):
    yield (('U', 0, 0, lambda_), qubit)


@std.gate()
def H(qubit):
    yield (('u2', 0, pi), qubit)


@std.gate()
def U(q, theta, phi, lambda_):
    if abs(theta) < EPS:
        yield (('P', phi + lambda_), q)
    else:
        yield (('P', lambda_), q)
        if abs(theta - pi / 2) < EPS:
            yield (('rfUnitary', pi / 2, pi / 2), q)
        else:
            yield (('rfUnitary', pi / 2, 0), q)
            yield (('rfUnitary', pi / 2, pi - theta), q)
            yield (('P', theta), q)
        yield (('P', phi), q)


@std.gate()
def X(q):
    yield (('u3', pi, 0, pi), q)


@std.gate()
def Y(q):
    yield (('u3', pi, pi / 2, pi / 2), q)


@std.gate()
def Z(q):
    yield (('u1', pi), q)


@std.gate()
def S(q):
    yield (('u1', pi / 2), q)


@std.gate(name='-S')
def Sdg(q):
    yield (('u1', -pi / 2), q)


@std.gate()
def T(q):
    yield (('u1', pi / 4), q)


@std.gate(name='-T')
def Tdg(q):
    yield (('u1', -pi / 4), q)


@std.gate(name='X/2')
def sx(q):
    yield ('-S', q)
    yield ('H', q)
    yield ('-S', q)


@std.gate(name='-X/2')
def sxdg(q):
    yield ('S', q)
    yield ('H', q)
    yield ('S', q)


@std.gate(name='Y/2')
def sy(q):
    yield ('Z', q)
    yield ('H', q)


@std.gate(name='-Y/2')
def sydg(q):
    yield ('H', q)
    yield ('Z', q)


@std.gate()
def Rx(q, theta):
    yield (('u3', theta, -pi / 2, pi / 2), q)


@std.gate()
def Ry(q, theta):
    yield (('u3', theta, 0, 0), q)


@std.gate(name='W/2')
def W2(q):
    yield (('u3', pi / 2, -pi / 4, pi / 4), q)


@std.gate(name='-W/2')
def W2(q):
    yield (('u3', -pi / 2, -pi / 4, pi / 4), q)


@std.gate(name='V/2')
def W2(q):
    yield (('u3', pi / 2, -3 * pi / 4, 3 * pi / 4), q)


@std.gate(name='-V/2')
def W2(q):
    yield (('u3', -pi / 2, -3 * pi / 4, 3 * pi / 4), q)


@std.gate()
def Rz(q, phi):
    yield (('u1', phi), q)


@std.gate(2)
def Cnot(qubits):
    c, t = qubits
    yield ('H', t)
    yield ('CZ', (c, t))
    yield ('H', t)


@std.gate(2)
def crz(qubits, lambda_):
    c, t = qubits

    yield (('u1', lambda_ / 2), t)
    yield ('Cnot', (c, t))
    yield (('u1', -lambda_ / 2), t)
    yield ('Cnot', (c, t))


@std.opaque('Delay')
def delay(ctx, qubits, time):
    qubit, = qubits
    yield ('!add', 'time', time), qubit


@std.opaque('P')
def P(ctx, qubits, phi):
    from ..compiler import call_opaque

    phi += ctx.phases[qubits[0]]
    yield ('!set', 'phase', 0), qubits[0]

    call_opaque((('rfUnitary', pi / 2, pi / 2), *qubits), ctx, std)
    call_opaque((('rfUnitary', phi, 0), *qubits), ctx, std)
    call_opaque((('rfUnitary', pi / 2, -pi / 2), *qubits), ctx, std)


@std.opaque('Barrier')
def barrier(ctx, qubits):
    time = max(ctx.time[qubit] for qubit in qubits)
    for qubit in qubits:
        yield ('!set', 'time', time), qubit


@std.opaque('rfPulse')
def rfPulse(ctx, qubits, waveform):
    qubit, = qubits

    if isinstance(waveform, str):
        waveform = wave_eval(waveform)

    yield ('!add', 'waveform', waveform >> ctx.time[qubit]), ('RF', qubit)


@std.opaque('fluxPulse')
def fluxPulse(ctx, qubits, waveform):
    qubit, = qubits

    if isinstance(waveform, str):
        waveform = wave_eval(waveform)

    yield ('!add', 'waveform', waveform >> ctx.time[qubit]), ('Z', qubit)


@std.opaque('Pulse')
def Pulse(ctx, qubits, channel, waveform):

    if isinstance(waveform, str):
        waveform = wave_eval(waveform)

    t = max(ctx.time[qubit] for qubit in qubits)

    yield ('!add', 'waveform', waveform >> t), (channel, *qubits)


@std.opaque('setBias')
def setBias(ctx, qubits, channel, bias, edge=0, buffer=0):
    yield ('!set', 'bias', (bias, edge, buffer)), (channel, *qubits)
    time = max(ctx.time[qubit] for qubit in qubits)
    for qubit in qubits:
        yield ('!set', 'time', time + buffer), qubit


@std.opaque('I')
def I(ctx, qubits):
    qubit = qubits[0]
    yield ('!add', 'waveform', zero()), ('RF', qubit)
    yield ('!add', 'waveform', zero()), ('Z', qubit)
    yield ('!add', 'waveform', zero()), ('readoutLine.RF', qubit)


def _rfUnitary(ctx, qubits, theta, phi, level1=0, level2=1):
    from numpy import interp

    qubit, = qubits

    if theta < 0:
        theta = -theta
        phi += pi
    theta = mod(theta, 2 * pi)
    if theta > pi:
        theta = 2 * pi - theta
        phi += pi

    phi = mod(
        phi + ctx.phases_ext[qubit][level1] - ctx.phases_ext[qubit][level2],
        2 * pi)
    phi = phi if abs(level2 - level1) % 2 else phi - pi
    if phi > pi:
        phi -= 2 * pi
    phi = phi / (level2 - level1)

    shape = ctx.params.get('shape', 'CosPulse')
    buffer = ctx.params.get('buffer', 0)
    alpha = ctx.params.get('alpha', 1)
    beta = ctx.params.get('beta', 0)
    delta = ctx.params.get('delta', 0)

    phase = pi * interp(phi / pi, *ctx.params['phase'])

    pulseLib = {
        'CosPulse': cosPulse,
        'Gaussian': gaussian,
        'Square': square,
        'cosPulse': cosPulse,
        'gaussian': gaussian,
        'square': square,
        'DC': square,
    }

    while theta > 0:
        if theta > pi / 2:
            theta1 = pi / 2
            theta -= pi / 2
        else:
            theta1 = theta
            theta = 0

        duration = interp(theta1 / pi, *ctx.params['duration'])
        amp = interp(theta1 / pi, *ctx.params['amp'])
        pulse = pulseLib[shape](duration)

        if duration > 0 and amp > 0:
            I, Q = mixing(amp * alpha * pulse,
                          phase=phase,
                          freq=delta,
                          DRAGScaling=beta / alpha)
            t = (duration + buffer) / 2 + ctx.time[qubit]
            if 'levels' in ctx.params:
                freq = (ctx.params['levels'][level2] -
                        ctx.params['levels'][level1]) / (level2 - level1)
            else:
                freq = ctx.params['frequency']
            wav, _ = mixing(I >> t, Q >> t, freq=freq)
            yield ('!add', 'waveform', wav), ('RF', qubit)
            yield ('!add', 'time', duration + buffer), qubit


@std.opaque('rfUnitary',
            params=[
                Parameter('shape', str, 'cosPulse'),
                Parameter('amp', list, [[0, 1], [0, 0.653]]),
                Parameter('duration', list, [[0, 1], [10e-9, 10e-9]]),
                Parameter('phase', list, [[-1, 1], [-1, 1]]),
                Parameter('frequency', float, 5e9, 'Hz'),
                Parameter('alpha', float, 1, 'Hz'),
                Parameter('beta', float, 0, 'Hz'),
                Parameter('delta', float, 0, 'Hz'),
                Parameter('buffer', float, 0, 's'),
            ])
def rfUnitary(ctx, qubits, theta, phi):
    yield from _rfUnitary(ctx, qubits, theta, phi)


@std.opaque('rfUnitary',
            type='BB1',
            params=[
                Parameter('shape', str, 'cosPulse'),
                Parameter('amp', list, [[0, 1], [0, 0.653]]),
                Parameter('duration', list, [[0, 1], [10e-9, 10e-9]]),
                Parameter('phase', list, [[-1, 1], [-1, 1]]),
                Parameter('frequency', float, 5e9, 'Hz'),
                Parameter('alpha', float, 1, 'Hz'),
                Parameter('beta', float, 0, 'Hz'),
                Parameter('delta', float, 0, 'Hz'),
                Parameter('buffer', float, 0, 's'),
            ])
def rfUnitary_BB1(ctx, qubits, theta, phi):
    import numpy as np

    p1 = np.arccos(-theta / (4 * pi))
    p2 = 3 * p1
    p1, p2 = p1 + phi, p2 + phi

    yield from _rfUnitary(ctx, qubits, pi, p1)
    yield from _rfUnitary(ctx, qubits, pi, p2)
    yield from _rfUnitary(ctx, qubits, pi, p2)
    yield from _rfUnitary(ctx, qubits, pi, p1)
    yield from _rfUnitary(ctx, qubits, theta, phi)


@std.opaque('Measure',
            params=[
                Parameter('duration', float, 1e-6, 's'),
                Parameter('amp', float, 0.1, 'a.u.'),
                Parameter('frequency', float, 6.5e9, 'Hz'),
                Parameter('bias', float, 0, 'a.u.'),
                Parameter('signal', str, 'state'),
                Parameter('weight', str, 'const(1)'),
                Parameter('phi', float, 0),
                Parameter('threshold', float, 0),
                Parameter('ring_up_amp', float, 0.1, 'a.u.'),
                Parameter('ring_up_time', float, 50e-9, 's')
            ])
def measure(ctx, qubits, cbit=None):
    from waveforms import cos, exp, pi, step

    qubit, = qubits

    if cbit is None:
        if len(ctx.measures) == 0:
            cbit = 0
        else:
            cbit = max(ctx.measures.keys()) + 1

    # lo = ctx.cfg._getReadoutADLO(qubit)
    amp = ctx.params['amp']
    duration = ctx.params['duration']
    frequency = ctx.params['frequency']
    bias = ctx.params.get('bias', None)
    signal = ctx.params.get('signal', 'state')
    ring_up_amp = ctx.params.get('ring_up_amp', amp)
    ring_up_time = ctx.params.get('ring_up_time', 50e-9)
    rsing_edge_time = ctx.params.get('rsing_edge_time', 5e-9)

    try:
        w = ctx.params['w']
        weight = None
    except:
        weight = ctx.params.get('weight',
                                f'square({duration}) >> {duration/2}')
        w = None

    t = ctx.time[qubit]

    # phi = 2 * np.pi * (lo - frequency) * t

    pulse = (ring_up_amp * (step(rsing_edge_time) >> t) - (ring_up_amp - amp) *
             (step(rsing_edge_time) >>
              (t + ring_up_time)) - amp * (step(rsing_edge_time) >>
                                           (t + duration)))
    yield ('!add', 'waveform',
           pulse * cos(2 * pi * frequency)), ('readoutLine.RF', qubit)
    if bias is not None:
        yield ('!set', 'bias', bias), ('Z', qubit)

    # pulse = square(2 * duration) >> duration
    # ctx.channel['readoutLine.AD.trigger', qubit] |= pulse.marker

    params = {k: v for k, v in ctx.params.items()}
    params['w'] = w
    params['weight'] = weight
    if cbit >= 0:
        yield ('!set', 'cbit',
               MeasurementTask(qubit, cbit, ctx.time[qubit], signal,
                               params)), cbit
    yield ('!set', 'time', t + duration), qubit
    yield ('!set', 'phase', 0), qubit


def parametric(ctx, qubits):
    t = max(ctx.time[q] for q in qubits)

    duration = ctx.params['duration']
    amp = ctx.params['amp']
    offset = ctx.params['offset']
    frequency = ctx.params['frequency']

    if duration > 0:
        pulse = square(duration) >> duration / 2
        pulse = offset * pulse + amp * pulse * sin(2 * pi * frequency)
        yield ('!add', 'waveform', pulse >> t), ('coupler.Z', *qubits)

    for qubit in qubits:
        yield ('!set', 'time', t + duration), qubit

    yield ('!add', 'phase', ctx.params['phi1']), qubits[0]
    yield ('!add', 'phase', ctx.params['phi2']), qubits[1]


@std.opaque('CZ',
            params=[
                Parameter('duration', float, 50e-9, 's'),
                Parameter('amp', float, 0.8, 'a.u.'),
                Parameter('phi1', float, 0),
                Parameter('phi2', float, 0)
            ])
def CZ(ctx, qubits):
    t = max(ctx.time[q] for q in qubits)

    duration = ctx.params['duration']
    amp = ctx.params['amp']

    if amp != 0 and duration > 0:
        pulse = amp * (cos(pi / duration) * square(duration)) >> duration / 2
        yield ('!add', 'waveform', pulse >> t), ('coupler.Z', *qubits)

    for qubit in qubits:
        yield ('!set', 'time', t + duration), qubit

    yield ('!add', 'phase', ctx.params['phi1']), qubits[0]
    yield ('!add', 'phase', ctx.params['phi2']), qubits[1]


@std.opaque('CZ',
            type='parametric',
            params=[
                Parameter('duration', float, 50e-9, 's'),
                Parameter('amp', float, 0.8, 'a.u.'),
                Parameter('offset', float, 0, 'a.u.'),
                Parameter('frequency', float, 0, 'Hz'),
                Parameter('phi1', float, 0),
                Parameter('phi2', float, 0)
            ])
def CZ(ctx, qubits):
    yield from parametric(ctx, qubits)


@std.opaque('iSWAP',
            type='parametric',
            params={
                Parameter('duration', float, 50e-9, 's'),
                Parameter('amp', float, 0.8, 'a.u.'),
                Parameter('offset', float, 0, 'a.u.'),
                Parameter('frequency', float, 0, 'Hz'),
                Parameter('phi1', float, 0),
                Parameter('phi2', float, 0)
            })
def iSWAP(ctx, qubits):
    yield from parametric(ctx, qubits)



import copy
import pickle
import sys
from typing import Any, Generator


def flattenDictIter(d: dict,
                    prefix: list = []
                    ) -> Generator[tuple[str, Any], None, None]:
    for k in d:
        if isinstance(d[k], dict):
            yield from flattenDictIter(d[k], prefix=[*prefix, k])
        else:
            yield '.'.join(prefix + [k]), d[k]


def flattenDict(d: dict[str, Any]) -> dict[str, Any]:
    return {k: v for k, v in flattenDictIter(d)}


def foldDict(d: dict[str, Any]) -> dict[str, Any]:
    ret = {}
    for k, v in d.items():
        keys = k.split('.')
        d = ret
        for key in keys[:-1]:
            if key not in d:
                d[key] = dict()
            d = d[key]
        d[keys[-1]] = v
    return ret


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton,
                                        cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class _NOTSET(metaclass=Singleton):
    __slots__ = ()

    def __repr__(self):
        return 'N/A'


class _UNKNOW(metaclass=Singleton):
    __slots__ = ()

    def __repr__(self) -> str:
        return "Unknow"


NOTSET = _NOTSET()
UNKNOW = _UNKNOW()


class Delete(metaclass=Singleton):
    __slots__ = ()

    def __repr__(self):
        return 'Delete'


DELETE = Delete()


class Update():
    __slots__ = ('o', 'n')

    def __init__(self, o, n):
        self.o = o
        self.n = n

    def __repr__(self):
        return f"Update: {self.o!r} ==> {self.n!r}"

    def __eq__(self, o: object) -> bool:
        return isinstance(o, Update) and self.n == o.n


class Create():
    __slots__ = ('n', 'replace')

    def __init__(self, n, replace=False):
        """
        Create a new node

        :param n: the new node
        :param replace: if True, replace the node if it already exists
        """
        self.n = n
        self.replace = replace

    def __repr__(self):
        if self.replace:
            return f"Replace: {self.n!r}"
        else:
            return f"Create: {self.n!r}"

    def __eq__(self, o: object) -> bool:
        return isinstance(
            o, Create) and self.n == o.n and self.replace == o.replace


def _eq(a, b):
    import numpy as np

    if isinstance(a, np.ndarray):
        return np.array_equal(a, np.asarray(b))

    try:
        return a == b
    except:
        pass
    if isinstance(a, (list, tuple)):
        return len(a) == len(b) and all(_eq(a[i], b[i]) for i in range(len(a)))
    if isinstance(a, dict):
        return set(a.keys()) == set(b.keys()) and all(
            _eq(a[k], b[k]) for k in a)

    try:
        return pickle.dumps(a) == pickle.dumps(b)
    except:
        return False


def diff(d1: dict, d2: dict) -> dict:
    """
    Compute the difference between two dictionaries

    Args:
        d1: the original dictionary
        d2: the new dictionary

    Returns:
        a dictionary containing the difference between d1 and d2
    """
    ret = {}
    for k in d2:
        if k in d1:
            if isinstance(d2[k], type(d1[k])) and _eq(d1[k], d2[k]):
                pass
            elif isinstance(d1[k], dict) and isinstance(d2[k], dict):
                ret[k] = diff(d1[k], d2[k])
            else:
                ret[k] = Update(d1[k], d2[k])
        else:
            ret[k] = Create(d2[k])
    for k in d1:
        if k not in d2:
            ret[k] = DELETE
    return ret


def patch(source, diff, in_place=False):
    """
    Patch a dictionary with a diff

    Args:
        source: the original dictionary
        diff: the diff
        in_place: if True, patch the source dictionary in place
    
    Returns:
        the patched dictionary
    """
    if in_place:
        ret = source
    else:
        ret = copy.copy(source)
    for k, v in diff.items():
        if isinstance(v, dict):
            ret[k] = patch(source[k], v, in_place=in_place)
        else:
            if isinstance(v, Update):
                ret[k] = v.n
            elif isinstance(v, Create):
                if v.replace or k not in ret:
                    ret[k] = v.n
                else:
                    update_tree(ret[k], v.n)
            elif v is DELETE:
                del ret[k]
            else:
                raise ValueError(f"Unsupported patch: {v!r}")
    return ret


def merge(diff1, diff2, origin=None):
    """
    Merge two diffs

    Args:
        diff1: the first diff
        diff2: the second diff
        origin: the original dictionary

    Returns:
        the merged diff
    """
    if origin is not None:
        updated = patch(patch(origin, diff1), diff2)
        return diff(origin, updated)

    ret = {}
    for k, v in diff1.items():
        if k in diff2:
            v2 = diff2[k]
            if isinstance(v, dict) and isinstance(v2, dict):
                d = merge(v, v2)
                if d:
                    ret[k] = d
            else:
                if isinstance(v, Update) and isinstance(v2, Update):
                    ret[k] = Update(v.o, v2.n)
                elif isinstance(v, Create) and isinstance(v2, dict):
                    ret[k] = Create(patch(copy.copy(v.n), v2, True))
                elif isinstance(v, Create) and isinstance(v2, Update):
                    ret[k] = Create(v2.n)
                elif isinstance(v, Create) and v2 is DELETE:
                    pass
                elif v2 is DELETE:
                    ret[k] = DELETE
                elif v is DELETE and isinstance(v2, Create):
                    if isinstance(v2.n, dict):
                        ret[k] = Create(v2.n, replace=True)
                    else:
                        ret[k] = Update(UNKNOW, v2.n)
                elif isinstance(v2, Create) and v2.replace:
                    ret[k] = v2
                else:
                    raise ValueError(f"Unsupported merge: {v!r} {v2!r}")
        else:
            ret[k] = v
    for k, v in diff2.items():
        if k not in diff1:
            ret[k] = v
    return ret


def print_diff(d, limit=None, offset=0, file=sys.stdout):
    """
    Print a diff

    Args:
        d: the diff
        limit: the maximum number of lines to print
        offset: the offset of the first line
        file: the file to print to
    """
    count = 0
    for i, (k, v) in enumerate(flattenDictIter(d)):
        if i >= offset:
            print(f"{k:40}", v, file=file)
            count += 1
            if limit is not None and count >= limit:
                break


def update_tree(result, updates):
    for k, v in updates.items():
        if isinstance(v, dict):
            if k not in result or not isinstance(result[k], dict):
                result[k] = {}
            update_tree(result[k], v)
        else:
            result[k] = v
    return result


def query_tree(q, dct, prefix=[]):
    keys = q.split('.')
    for i, key in enumerate(keys):
        if key not in dct:
            return (NOTSET, '.'.join(prefix + keys[:i + 1]))
        dct = dct[key]
    return dct


def sorted_tree(dct, *, keys=None):
    if keys is None or callable(keys):
        key = keys
    elif isinstance(keys, list):
        key = keys[0]
        if len(keys) > 1:
            keys = keys[1:]
        else:
            keys = None
    elif (isinstance(keys, tuple) and len(keys) == 2
          and isinstance(keys[1], dict)):
        key = keys[0]
        keys = keys[1]
    else:
        raise Exception(f"Unsupported keys: {keys!r}")

    if isinstance(dct, dict):
        if isinstance(keys, dict):
            default = keys.get('default', None)
            return {
                k: sorted_tree(dct[k], keys=keys.get(k, default))
                for k in sorted(dct.keys(), key=key)
            }
        else:
            return {
                k: sorted_tree(dct[k], keys=keys)
                for k in sorted(dct.keys(), key=key)
            }
    elif isinstance(dct, set):
        if isinstance(keys, dict):
            default = keys.get('default', None)
            return set([
                sorted_tree(v, keys=keys.get(v, default))
                for v in sorted(list(dct), key=key)
            ])
        else:
            return set([
                sorted_tree(v, keys=keys) for v in sorted(list(dct), key=key)
            ])
    else:
        return dct



from __future__ import annotations

import copy
import json
import random
from pathlib import Path
from typing import Any, Optional, Union

from .dicttree import flattenDict as _flattenDict
from .dicttree import flattenDictIter as _flattenDictIter
from .dicttree import foldDict as _foldDict
from .dicttree import update_tree as _update


def _query(q, dct):
    return {
        k.removeprefix(q + '.'): v
        for k, v in dct.items() if k.startswith(q + '.')
    }


def randomStr(n):
    s = ('abcdefghijklmnopqrstuvwxyz'
         'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
         '0123456789')
    return ''.join(random.choices(s, k=n))


def mixin(o: object, *traits: type) -> object:
    bases = (o.__class__, *traits)
    name = '_'.join([cls.__name__ for cls in bases])
    name = '_'.join([name, randomStr(6)])
    cls = type(name, bases, {})
    o.__class__ = cls
    return o


class TraitMeta(type):
    _traits = {}

    def __new__(cls, name, bases, namespace):
        cls = super().__new__(cls, name, bases, namespace)
        if name != 'Trait':
            TraitMeta._traits[name] = cls
        return cls


class Trait(metaclass=TraitMeta):
    pass


def queryKey(q, dct, prefix=None):
    if prefix is None:
        prefix = []

    keys = q.split('.', maxsplit=1)

    if not isinstance(dct, dict):
        k = '.'.join(prefix)
        raise KeyError(
            f"Query {k}.{q} error, type '{k}' is {type(dct)}, not dict.")
    try:
        sub = dct[keys[0]]
    except KeyError:
        k = '.'.join([*prefix, keys[0]])
        raise KeyError(
            f"Query {'.'.join([*prefix, q])} error, key '{k}' not found.")

    if len(keys) == 1:
        return sub

    else:
        return queryKey(keys[1], sub, [*prefix, keys[0]])


def query(q, dct, prefix=None):
    if isinstance(q, str):
        return queryKey(q, dct, prefix)
    elif isinstance(q, list):
        return [query(sub_q, dct, prefix) for sub_q in q]
    elif isinstance(q, tuple):
        return tuple([query(sub_q, dct, prefix) for sub_q in q])
    elif isinstance(q, set):
        return {sub_q: query(sub_q, dct, prefix) for sub_q in q}
    elif isinstance(q, dict):
        if prefix is None:
            prefix = []
        ret = {}
        for k, sub_q in q.items():
            if sub_q is None:
                ret[k] = queryKey(k, dct, prefix)
            else:
                ret[k] = query(sub_q, queryKey(k, dct, prefix), [*prefix, k])
        return ret
    else:
        raise TypeError


def setKey(q, value, dct, prefix=None):
    if prefix is None:
        prefix = []

    keys = q.split('.', maxsplit=1)

    if len(keys) == 1:
        if keys[0] in dct and isinstance(dct[keys[0]],
                                         dict) and not isinstance(value, dict):
            k = '.'.join([*prefix, keys[0]])
            raise ValueError(f'try to set a dict {k} to {type(value)}')
        else:
            dct[keys[0]] = value
    else:
        if keys[0] in dct and isinstance(dct[keys[0]], dict):
            sub = dct[keys[0]]
        elif keys[0] in dct and not isinstance(dct[keys[0]], dict):
            k = '.'.join([*prefix, keys[0]])
            raise ValueError(f'try to set a dict {k} to {type(value)}')
        else:
            sub = {}
            dct[keys[0]] = sub
        setKey(keys[1], sub, [*prefix, keys[0]])


class ConfigSection(dict):
    def __init__(self, cfg: BaseConfig, key: str):
        self._cfg_ = cfg
        self._key_ = key

    def __setitem__(self, key: str, value: Any) -> None:
        if self._cfg_ is None:
            self._modified_ = True
        else:
            self._cfg_._modified_ = True
        if isinstance(value, dict) and not isinstance(value, ConfigSection):
            key, *traits = key.split(':')
            if self._cfg_ is None:
                cfg = self
                k = key
            else:
                cfg = self._cfg_
                k = '.'.join([self._key_, key])
            d = ConfigSection(cfg, k)
            d.update(value)
            value = d
        elif isinstance(value, ConfigSection):
            value.__class__ = ConfigSection
        super().__setitem__(key, value)

    def __getitem__(self, key: str) -> ValueType:
        key, *traits = key.split(':')
        if self._cfg_ is not None:
            d = self._cfg_.query(self._key_)
            if self is not d:
                self.update(d)
        ret = super().__getitem__(key)
        if isinstance(ret, ConfigSection) and len(traits) > 0:
            traits = [
                TraitMeta._traits[n] for n in traits if n in TraitMeta._traits
            ]
            mixin(ret, *traits)
        return ret

    def __delitem__(self, key: str) -> None:
        if self._cfg_ is None:
            self._modified_ = True
        else:
            self._cfg_._modified_ = True
        return super().__delitem__(key)

    def __delattr__(self, name: str) -> None:
        if name in self:
            return self.__delitem__(name)
        else:
            return super().__delattr__(name)

    def __getattr__(self, name: str) -> Any:
        try:
            return self.__getattribute__(name)
        except:
            try:
                return self.__getitem__(name)
            except:
                raise AttributeError(f'Not Find Attr: {name}')

    def __setattr__(self, name: str, value: Any) -> None:
        if name in self:
            self.__setitem__(name, value)
        else:
            super().__setattr__(name, value)

    def __deepcopy__(self, memo):
        dct = {k: copy.deepcopy(v) for k, v in self.items()}
        return dct

    def query(self, q: Union[str, set, tuple, list,
                             dict]) -> Union[dict, ValueType]:
        if self._key_ is None:
            prefix = []
        else:
            prefix = self._key_.split('.')
        return query(q, self, prefix=prefix)

    def set(self, q, value):
        if self._key_ is None:
            prefix = []
        else:
            prefix = self._key_.split('.')
        setKey(q, value, self, prefix=prefix)

    def update(self, other):
        _update(self, other)


ValueType = Union[str, int, float, list, ConfigSection]


class BaseConfig(ConfigSection):
    def __init__(self,
                 path: Optional[Union[str, Path]] = None,
                 backup: bool = False):
        super().__init__(None, None)
        if isinstance(path, str):
            path = Path(path)
        self._path_ = path
        self._backup_ = backup
        self._modified_ = False

        if self._path_ is None:
            return
        if self._path_.exists():
            self.reload()
            if '__version__' not in self:
                self['__version__'] = 1
        else:
            self._path_.parent.mkdir(parents=True, exist_ok=True)
            self['__version__'] = 1
            self._modified_ = True
            self.commit()

    def commit(self):
        if not self._modified_ or self._path_ is None:
            return
        if self._backup_ and self._path_.exists():
            v = self['__version__']
            bk = self._path_.with_stem(self._path_.stem + f"_{v}")
            self._path_.rename(bk)
        with self._path_.open('w') as f:
            self['__version__'] = self['__version__'] + 1
            json.dump(self, f, indent=4)
            self._modified_ = False

    def rollback(self):
        if not self._modified_ or self._path_ is None:
            return
        self.reload()

    def reload(self):
        with self._path_.open('r') as f:
            dct = json.load(f)
            self.clear()
            self.update(dct)
            self._modified_ = False

    @classmethod
    def fromdict(cls, d: dict) -> BaseConfig:
        ret = cls()
        ret.update(d)
        return ret



from numpy import e, pi

from .cache import cache
from .quantum import (Config, compile, getConfig, libraries, qpt, qptInitList,
                      qst, qst_mle, qstOpList, setConfig, stdlib)
from .quantum.transmon import Transmon
from .version import __version__
from .waveform import (D, Waveform, chirp, const, cos, cosPulse, cut, exp,
                       function, gaussian, general_cosine, hanning, interp,
                       mixing, one, poly, registerBaseFunc, registerDerivative,
                       samplingPoints, sign, sin, sinc, square, step, zero)
from .waveform_parser import wave_eval



from ..dicttree import NOTSET
from .arch import Architecture, get_arch, register_arch
from .base import (ABCCompileConfigMixin, ADChannel, AWGChannel,
                   MeasurementTask, MultADChannel, MultAWGChannel, QLispCode,
                   Signal)
from .commands import (COMMAND, FREE, PUSH, READ, SYNC, TRIG, WRITE,
                       CommandList, DataMap, RawData, Result)
from .compiler import compile
from .config import Config, ConfigProxy, GateConfig
from .library import Library, Parameter, libraries
from .libs import std as stdlib
from .prog import Program, ProgramFrame



from __future__ import annotations

import re
from typing import Any, NamedTuple


class Token(NamedTuple):
    type: str
    value: str
    line: int
    column: int


class Symbol():
    __slots__ = ('name', )

    def __init__(self, name):
        self.name = name

    def __hash__(self):
        return hash(self.name)

    def __repr__(self):
        if isinstance(self.name, str):
            return self.name
        return repr(self.name)

    def __eq__(self, other):
        return isinstance(other, Symbol) and other.name == self.name


class Expression(tuple):
    pass


def atom(token):
    "Numbers become numbers; every other token is a symbol."
    for kind in [int, float, complex, Symbol]:
        try:
            return kind(token)
        except ValueError:
            continue
    raise ValueError()


Number = (int, float, complex
          )  # A Lisp Number is implemented as a Python int or float


def tokenize(code):
    keywords = {}
    token_specification = [
        ('BRACKET', r'[\(\)]'),
        ('STRING', r'\"([^\\\"]|\\.)*\"'),
        ('NEWLINE', r'\n'),  # Line endings
        ('SKIP', r'[ \t]+'),  # Skip over spaces and tabs
        ('ATOM', r'[A-Za-z0-9\.!@#$%^&\*/_\-\+:\?\|\<\>=]+'),  # Atom
        ('COMMENT', r';'),  # Comment
        ('QUOTE', r"'"),
        ('MISMATCH', r'.'),  # Any other character
    ]
    tok_regex = '|'.join('(?P<%s>%s)' % pair for pair in token_specification)
    line_num = 1
    line_start = 0
    in_comment = False
    for mo in re.finditer(tok_regex, code):
        kind = mo.lastgroup
        value = mo.group()
        column = mo.start() - line_start

        if kind == 'NEWLINE':
            line_start = mo.end()
            line_num += 1
            in_comment = False
            continue
        elif kind == 'COMMENT':
            in_comment = True
            continue
        elif in_comment or kind == 'SKIP':
            continue
        elif kind == 'STRING':
            value = value[1:-1]
        elif kind == 'BRACKET':
            kind = value
        elif kind == 'ATOM':
            value = atom(value)
        elif kind == 'MISMATCH':
            raise RuntimeError(f'{value!r} unexpected on line {line_num}')
        yield Token(kind, value, line_num, column)



import tempfile
from ast import literal_eval
from functools import lru_cache

import ply.lex as lex
import ply.yacc as yacc

from . import waveform


class _WaveLexer:
    """Waveform Lexer.
    """

    def __init__(self):
        """Create a PLY lexer."""
        self.lexer = lex.lex(module=self, debug=False)

    def input(self, data):
        self.lexer.input(data)

    def token(self):
        """Return the next token."""
        ret = self.lexer.token()
        return ret

    literals = r'=()[]<>,.+-/*^'
    functions = [
        'D', 'chirp', 'const', 'cos', 'cosPulse', 'cut', 'exp', 'gaussian',
        'general_cosine', 'hanning', 'interp', 'mixing', 'one', 'poly',
        'samplingPoints', 'sign', 'sin', 'sinc', 'square', 'step', 'zero'
    ]
    tokens = [
        'REAL', 'IMAG', 'INT', 'STRING', 'ID', 'LSHIFT', 'RSHIFT', 'POW',
        'CONST', 'FUNCTION'
    ]

    def t_ID(self, t):
        r'[a-zA-Z_][a-zA-Z_0-9]*'
        if t.value in ['pi', 'e', 'inf']:
            t.type = 'CONST'
            return t
        if t.value in self.functions:
            t.type = 'FUNCTION'
            return t
        else:
            return t

    def t_IMAG(self, t):
        r'((([0-9]+|([0-9]+)?\.[0-9]+|[0-9]+\.)[eE][+-]?[0-9]+)|(([0-9]+)?\.[0-9]+|[0-9]+\.)|[1-9][0-9]*|0)j'
        return t

    def t_REAL(self, t):
        r'(([0-9]+|([0-9]+)?\.[0-9]+|[0-9]+\.)[eE][+-]?[0-9]+)|(([0-9]+)?\.[0-9]+|[0-9]+\.)'
        return t

    def t_INT(self, t):
        r'[1-9][0-9]*|0'
        return t

    def t_STRING(self, t):
        r'(".*")|(\'.*\')'
        return t

    def t_LSHIFT(self, t):
        '<<'
        return t

    def t_RSHIFT(self, t):
        '>>'
        return t

    def t_POW(self, t):
        r'\*\*'
        return t

    def t_eof(self, _):
        return None

    t_ignore = ' \t\r\n'

    def t_error(self, t):
        raise SyntaxError("Unable to match any token rule, got -->%s<-- " %
                          t.value)


class _WaveParser:

    def __init__(self):
        self.lexer = _WaveLexer()
        self.tokens = self.lexer.tokens
        self.parse_dir = tempfile.mkdtemp(prefix='waveforms')
        self.precedence = (('left', 'RSHIFT', 'LSHIFT'), ('left', '+', '-'),
                           ('left', '*', '/'), ('left', 'POW',
                                                '^'), ('right', 'UMINUS'))
        self.parser = yacc.yacc(module=self,
                                debug=False,
                                outputdir=self.parse_dir)
        self.waveform = None

    def parse(self, data):
        #self.waveform = None
        self.parser.parse(data, lexer=self.lexer, debug=False)
        if self.waveform is None:
            raise SyntaxError("Uncaught exception in parser; " +
                              "see previous messages for details.")
        if isinstance(self.waveform, (float, int)):
            self.waveform = waveform.const(self.waveform)
        return self.waveform.simplify()

    def getFunction(self, name):
        return getattr(waveform, name)

    # ---- Begin the PLY parser ----
    start = 'main'

    def p_main(self, p):
        """
        main : expression
        """
        self.waveform = p[1]

    def p_const(self, p):
        """
        expression : CONST
        """
        p[0] = {'pi': waveform.pi, 'e': waveform.e, 'inf': waveform.inf}[p[1]]

    def p_real_imag_int_string(self, p):
        """
        expression : REAL
                   | IMAG
                   | INT
                   | STRING
        """
        p[0] = literal_eval(p[1])

    def p_tuple_list(self, p):
        """
        expression : tuple
                   | list
        """
        p[0] = p[1]

    def p_expr_uminus(self, p):
        """
        expression : '-' expression %prec UMINUS
        """
        p[0] = -p[2]

    def p_function_call(self, p):
        """
        expression : FUNCTION '(' ')'
        """
        p[0] = self.getFunction(p[1])()

    def p_function_call_2(self, p):
        """
        expression :  FUNCTION '(' args ')'
        """
        p[0] = self.getFunction(p[1])(*p[3])

    def p_function_call_3(self, p):
        """
        expression :  FUNCTION '(' kwds ')'
        """
        p[0] = self.getFunction(p[1])(**p[3])

    def p_function_call_4(self, p):
        """
        expression :  FUNCTION '(' args ',' kwds ')'
        """
        p[0] = self.getFunction(p[1])(*p[3], **p[5])

    def p_bracket(self, p):
        """
        expression :  '(' expression ')'
        """
        p[0] = p[2]

    def p_binary_operators(self, p):
        """
        expression : expression '+' expression
                   | expression '-' expression
                   | expression '*' expression
                   | expression '/' expression
                   | expression LSHIFT expression
                   | expression RSHIFT expression
                   | expression '^' expression
                   | expression POW expression
        """
        if p[2] == '+':
            p[0] = p[1] + p[3]
        elif p[2] == '-':
            p[0] = p[1] - p[3]
        elif p[2] == '*':
            p[0] = p[1] * p[3]
        elif p[2] == '/':
            p[0] = p[1] / p[3]
        elif p[2] == '>>':
            p[0] = p[1] >> p[3]
        elif p[2] == '<<':
            p[0] = p[1] << p[3]
        elif p[2] == '^':
            p[0] = p[1]**p[3]
        else:
            p[0] = p[1]**p[3]

    def p_expr_list_2(self, p):
        """
        expr_list : expression ',' expression
        """
        p[0] = [p[1], p[3]]

    def p_expr_list_3(self, p):
        """
        expr_list : expr_list ',' expression
        """
        p[0] = [*p[1], p[3]]

    def p_tuple(self, p):
        """
        tuple : '(' expression ',' ')'
              | '(' expr_list ')'
        """
        if len(p) == 5:
            p[0] = (p[2], )
        else:
            p[0] = tuple(p[2])

    def p_list_1(self, p):
        """
        list : '[' expression ']'
        """
        p[0] = [p[2]]

    def p_list_2(self, p):
        """
        list : '[' expr_list ']'
        """
        p[0] = p[2]

    def p_args(self, p):
        """
        args : expression
             | args ',' expression
        """
        if len(p) == 2:
            p[0] = (p[1], )
        else:
            p[0] = p[1] + (p[3], )

    def p_kwds(self, p):
        """
        kwds : ID '=' expression
             | kwds ',' ID '=' expression
        """
        if len(p) == 4:
            p[0] = {p[1]: p[3]}
        else:
            kwds = {}
            kwds.update(p[1])
            kwds[p[3]] = p[5]
            p[0] = kwds
            # p[0] = p[1] | {p[3]: p[5]}   # only works on Python>=3.9

    def p_error(self, p):
        raise SyntaxError("Syntax error in input!")


_wave_parser = _WaveParser()


@lru_cache(maxsize=1024)
def wave_eval(expr: str) -> waveform.Waveform:
    try:
        return _wave_parser.parse(expr)
    except:
        raise SyntaxError(f"Illegal expression '{expr}'")



from importlib import reload, import_module
from functools import wraps


class AutoReload():

    def __new__(cls, *args, **kwargs):
        mod = reload(import_module(cls.__module__))
        cls = getattr(mod, cls.__name__)
        return super().__new__(cls, *args, **kwargs)


def autoreload(func):

    module_name = func.__module__
    func_name = func.__name__

    @wraps(func)
    def wrapper(*args, **kwargs):
        mod = reload(import_module(module_name))
        func = getattr(mod, func_name)
        return func._func_(*args, **kwargs)

    wrapper._func_ = func

    return wrapper


from itertools import chain
from typing import Optional

from ..base import QLispError
from ..library import Library
from .node import *
from .qasm import Qasm

__qindex = 0
__cindex = 0


def allocaInit():
    global __qindex, __cindex
    __qindex, __cindex = 0, 0


def allocaQubits(size):
    global __qindex
    addresses = tuple(range(__qindex, __qindex + size))
    __qindex += size
    return addresses


def allocaCbits(size):
    global __cindex
    addresses = tuple(range(__cindex, __cindex + size))
    __cindex += size
    return addresses


def get_name(name, scope):
    for s in scope[::-1]:
        if name in s:
            return s[name]
    else:
        raise QLispError(f'Undefined {name:r}')


def get_sym(ID, scope, index=None):
    """Return the correspond symbolic number."""
    if not scope:
        if index is not None:
            return ID[index]
        else:
            return ID
    if ID.name not in scope[-1]:
        raise QLispError(
            "Expected local parameter name: ",
            "name=%s, line=%s, file=%s" % (ID.name, ID.line, ID.file))
    if isinstance(ID, IndexedId):
        return get_sym(scope[-1][ID.name], scope[0:-1], ID.index)
    else:
        return get_sym(scope[-1][ID.name], scope[0:-1])


def qasm_eval_single_opaque(st, opaque, scope):
    assert len(opaque.bitlist.children) == len(st.bitlist.children)
    bitlist = [get_sym(q, scope) for q in st.bitlist.children]
    if st.arguments is not None:
        assert len(opaque.arguments.children) == len(st.arguments.children)
        args = [a.real(scope) for a in st.arguments.children]
        gate = (st.name, *args)
    else:
        gate = st.name
    if isinstance(bitlist[0], int):
        yield (gate, tuple(bitlist) if len(bitlist) > 1 else bitlist[0])
    else:
        yield from ((gate, qubits if len(qubits) > 1 else qubits[0])
                    for qubits in zip(*bitlist))


def qasm_eval_single_command(st, scope):
    if isinstance(st, Measure):
        q = get_sym(st.children[0], scope)
        c = get_sym(st.children[1], scope)
        if isinstance(q, tuple):
            assert len(q) == len(c)
            yield from ((('Measure', c_), q_) for c_, q_ in zip(c, q))
        else:
            yield (('Measure', c), q)
    elif isinstance(st, Barrier):
        q = tuple(chain(*[get_sym(_, scope) for _ in st.children[0].children]))
        yield ('Barrier', q)
    elif isinstance(st, Reset):
        q = get_sym(st.children[0], scope)
        if isinstance(q, tuple):
            yield from (('Reset', q_) for q_ in q)
        else:
            yield ('Reset', q)


def qasm_eval_if(st, scope):
    # TODO
    yield st


def qasm_eval_prog(prog, scope=None):
    current_scope = {}
    if scope is None:
        allocaInit()
        scope = [current_scope]

    for st in prog.children:
        if isinstance(st, (Gate, Opaque)):
            current_scope[st.name] = st
        elif isinstance(st, Qreg):
            current_scope[st.name] = allocaQubits(st.index)
        elif isinstance(st, Creg):
            current_scope[st.name] = allocaCbits(st.index)
        elif isinstance(st, CustomUnitary):
            gate = get_name(st.name, scope)
            sub_scope = {}
            assert len(gate.bitlist.children) == len(st.bitlist.children)
            for name, q in zip(gate.bitlist.children, st.bitlist.children):
                sub_scope[name.name] = q
            if gate.arguments is not None:
                assert len(gate.arguments.children) == len(
                    st.arguments.children)
                for name, a in zip(gate.arguments.children,
                                   st.arguments.children):
                    sub_scope[name.name] = a
            if isinstance(gate, Opaque):
                yield from qasm_eval_single_opaque(st, gate, scope)
            elif isinstance(gate, Gate):
                yield from qasm_eval_prog(gate.body, [*scope, sub_scope])
            else:
                raise QLispError(f"{st.name:r} is not gate nor opaque")
        elif isinstance(st, If):
            yield from qasm_eval_if(st, scope)
        elif isinstance(st, Format):
            pass
        else:
            yield from qasm_eval_single_command(st, scope)


def qasm_eval(data: str, lib: Optional[Library] = None):
    qasm = Qasm(data=data, lib=lib)
    prog = qasm.parse()
    yield from qasm_eval_prog(prog)



from pathlib import Path
import pickle
import struct
from itertools import chain, product

import numpy as np

from ...cache import cache_dir
from .mat import mat2num, num2mat
from .seq2mat import seq2mat

one_qubit_clifford_seq = [
    # Paulis
    ('I', ),
    ('X', ),
    ('Y', ),
    ('Y', 'X'),

    # 2 pi / 3 rotations
    ('X/2', 'Y/2'),
    ('X/2', '-Y/2'),
    ('-X/2', 'Y/2'),
    ('-X/2', '-Y/2'),
    ('Y/2', 'X/2'),
    ('Y/2', '-X/2'),
    ('-Y/2', 'X/2'),
    ('-Y/2', '-X/2'),

    # pi / 2 rotations
    ('X/2', ),
    ('-X/2', ),
    ('Y/2', ),
    ('-Y/2', ),
    ('-X/2', 'Y/2', 'X/2'),
    ('-X/2', '-Y/2', 'X/2'),

    # Hadamard-like
    ('X', 'Y/2'),
    ('X', '-Y/2'), # Hadamard
    ('Y', 'X/2'),
    ('Y', '-X/2'),
    ('X/2', 'Y/2', 'X/2'),
    ('-X/2', 'Y/2', '-X/2')
] #yapf: disable


one_qubit_clifford_seq_2 = [
    # Paulis
    ('I',),
    ('X',),
    ('Y',),
    ('Z',),

    # 2 pi / 3 rotations
    ('Y/2', '-S'),
    ('-Y/2', 'S'),
    ('Y/2', 'S'),
    ('-Y/2', '-S'),
    ('X/2', 'S'),
    ('-X/2', '-S'),
    ('X/2', '-S'),
    ('-X/2', 'S'),

    # pi / 2 rotations
    ('X/2',),
    ('-X/2',),
    ('Y/2',),
    ('-Y/2',),
    ('S',),
    ('-S',),

    # Hadamard-like
    ('Y/2', 'Z'),
    ('-Y/2', 'Z'), # Hadamard
    ('X/2', 'Z'),
    ('-X/2', 'Z'),
    ('X', 'S'),
    ('Y', 'S')
] #yapf: disable


def _fill_seq_pair(A, B):
    """
    填充必要的 I 作为占位符
    """
    if len(A) > len(B):
        B = ('I', ) * (len(A) - len(B)) + B
    elif len(B) > len(A):
        A = ('I', ) * (len(B) - len(A)) + A
    return A, B


def _short_seq_pair(A, B):
    """
    去掉多余的 I
    """
    C, D = [], []
    for x, y in zip(A, B):
        if x == 'I' and y == 'I':
            continue
        else:
            C.append(x)
            D.append(y)
    if len(C) == 0:
        return ('I', ), ('I', )
    else:
        return tuple(C), tuple(D)


def generateTwoQubitCliffordSequence():
    """
    生成全部 11520 个群元对应的操作序列
    """
    C1 = one_qubit_clifford_seq
    S1 = [0, 8, 7]  #  I, Rot(2/3 pi, (1,1,1)), Rot(2/3 pi, (1,-1,-1))

    # single qubit class 0 ~ 576
    for i, j in product(range(24), repeat=2):
        A, B = _fill_seq_pair(C1[i], C1[j])
        yield _short_seq_pair(A, B)

    # CNOT-like class 576 ~ 5760
    for i, j in product(range(24), repeat=2):
        for k, l in product(S1, S1):
            A, B = _fill_seq_pair(C1[i], C1[j])
            C, D = _fill_seq_pair(C1[k], C1[l])
            yield _short_seq_pair(
                A + ('CX', ) + C,
                B + ('CX', ) + D,
            )

    # iSWAP-like class 5760 ~ 10944
    for i, j in product(range(24), repeat=2):
        for k, l in product(S1, S1):
            A, B = _fill_seq_pair(C1[i], C1[j])
            C, D = _fill_seq_pair(C1[k], C1[l])
            yield _short_seq_pair(A + ('iSWAP', ) + C, B + ('iSWAP', ) + D)

    # SWAP-like class 10944 ~ 11520
    for i, j in product(range(24), repeat=2):
        A, B = _fill_seq_pair(C1[i], C1[j])
        yield _short_seq_pair(A + ('SWAP', ), B + ('SWAP', ))


def two_qubit_clifford_num():
    return [
        mat2num(seq2mat(seq)) for seq in generateTwoQubitCliffordSequence()
    ]


NUMBEROFELEMENTS = 11520  # cliffordOrder(2)

__base = cache_dir.parent / "clifford"
__mul_table_file = __base / "clifford_2qb_mul_table_unsigned_short.dat"
__index_file = __base / "clifford_2qb_index.pickle"
__seq_lib_file = __base / "clifford_2qb_seq_lib.db"
__mul_table_packer = struct.Struct('<H')

if not __base.exists():
    __base.mkdir(parents=True)

if not __mul_table_file.exists():
    __mul_table = bytearray(NUMBEROFELEMENTS * NUMBEROFELEMENTS * 2)
    for offset in range(0, len(__mul_table), 2):
        __mul_table_packer.pack_into(__mul_table, offset, 0x8000)
    __mul_table_file.write_bytes(__mul_table)
else:
    __mul_table = bytearray(__mul_table_file.read_bytes())

if not __index_file.exists():
    __index2num = two_qubit_clifford_num()
    __num2index = {n: i for i, n in enumerate(__index2num)}
    __index2mat = [num2mat(n) for n in __index2num]

    with open(__index_file, 'wb') as f:
        pickle.dump((__index2num, __num2index, __index2mat), f)

else:
    with open(__index_file, 'rb') as f:
        (__index2num, __num2index, __index2mat) = pickle.load(f)

# if not __seq_lib_file.exists():
#     __index2seq = [{seq} for seq in generateTwoQubitCliffordSequence()]
#     __seq2index = {
#         seq: i
#         for i, seq in enumerate(generateTwoQubitCliffordSequence())
#     }
#     with open(__seq_lib_file, 'wb') as f:
#         pickle.dump(__index2seq, __seq2index, f)
# else:
#     with open(__seq_lib_file, 'rb') as f:
#         __index2seq, __seq2index = pickle.load(f)

# def _elms(seq):
#     base = {'I'}
#     for g1, g2 in zip(*seq):
#         if (g1, g2) in [('CZ', 'CZ'), ('iSWAP', 'iSWAP'),
#                         ('SQiSWAP', 'SQiSWAP')]:
#             base.add(g1)
#         elif (g1, g2) in [('C', 'Z'), ('Z', 'C')]:
#             base.add('CZ')
#         elif g1 == 'C' or g2 == 'C':
#             base.add(g1 + g2)
#         else:
#             base.add(g1)
#             base.add(g2)
#     return base

# def _betterSeq(a, b):
#     """
#     a is better than b
#     """
#     return (len(a[0]) < len(b[0]) and _elms(a) <= _elms(b)
#             or len(a[0]) == len(b[0]) and _elms(a) < _elms(b))

# def _updateSeqLib(i, seq):
#     if all(_betterSeq(s, seq) for s in __index2seq[i]):
#         return
#     l = set()
#     for s in __index2seq[i]:
#         if _betterSeq(seq, s):
#             continue
#         l.add(s)
#     l.add(seq)
#     __index2seq[i] = l
#     with open(__seq_lib_file, 'wb') as f:
#         pickle.dump(__index2seq, f)


def mat2index(mat: np.ndarray) -> int:
    """
    convert matrix to index

    Args:
        mat ([type]): unitary matrix

    Returns:
        int: index of Clifford gate
    """
    return num2index(mat2num(mat))


def index2mat(i: int) -> np.ndarray:
    """
    convert index to matrix

    Args:
        i (int): index of Clifford gate

    Returns:
        np.ndarray: matrix
    """
    return __index2mat[i]


def index2num(i: int) -> int:
    return __index2num[i]


def num2index(num: int) -> int:
    return __num2index[num]


def seq2num(seq):
    return mat2num(seq2mat(seq))


def seq2index(seq):
    # if seq in __seq2index:
    #     return __seq2index[seq]
    i = mat2index(seq2mat(seq))
    # _updateSeqLib(i, seq)
    return i


# def index2seq(i, base=None):
#     if base is None:
#         return __index2seq[i]
#     ret = []
#     for seq in __index2seq[i]:
#         if _elms(seq) <= set(base) | {'I'}:
#             ret.append(seq)
#     return ret


def mul(i1: int, i2: int) -> int:
    offset = 2 * (NUMBEROFELEMENTS * i1 + i2)
    ret = __mul_table_packer.unpack_from(__mul_table, offset)[0]

    if ret & 0x8000:
        ret = mat2index(index2mat(i1) @ index2mat(i2))
        __mul_table_packer.pack_into(__mul_table, offset, ret)
        with open(__mul_table_file, 'r+b') as f:
            f.seek(offset)
            f.write(__mul_table_packer.pack(ret))
    return ret


def inv(i: int) -> int:
    return mat2index(index2mat(i).T.conj())


def _genSeq(i, gate=('CZ', 'CZ')):
    pulses = ['I', 'X', 'Y', 'X/2', 'Y/2', '-X/2', '-Y/2']
    phases = ['I', 'Z', 'S', '-S']

    for k in [2, 4, 6, 8]:
        num = len(pulses)**k * len(phases)**2
        if i >= num:
            i -= num
            continue
        index = np.unravel_index(i, (len(pulses), ) * k + (len(phases), ) * 2)
        a = [pulses[n] for n in index[:-2]]
        p = [phases[n] for n in index[-2:]]
        return _short_seq_pair(
            tuple(
                chain(*[[a[j], gate[0]] for j in range(0, k - 3, 2)],
                      [a[-2], p[0]])),
            tuple(
                chain(*[[a[j], gate[1]] for j in range(1, k - 2, 2)],
                      [a[-1], p[1]])))
    else:
        raise IndexError(f'i={i} should be less than 94158400.')


def _countTwoQubitGate(seq, gate):
    count = 0
    for a, b in zip(*seq):
        if (a, b) == gate:
            count += 1
    return count


def genSeqForGate(db, gate=('CZ', 'CZ')):
    db = Path(db)
    if db.exists():
        with open(db, 'rb') as f:
            start, index2seq = pickle.load(f)
    else:
        start, index2seq = 0, [list() for i in range(NUMBEROFELEMENTS)]
    try:
        while True:
            try:
                seq = _genSeq(start, gate)
            except IndexError:
                break
            i = seq2index(seq)
            if len(index2seq[i]) == 0:
                index2seq[i].append(seq)
            else:
                s = index2seq[i].pop()
                if (len(s[0]) == len(seq[0])
                        and _countTwoQubitGate(s, gate) == _countTwoQubitGate(
                            seq, gate)):
                    index2seq[i].append(seq)
                    index2seq[i].append(s)
                elif (len(s[0]) > len(seq[0])
                      or _countTwoQubitGate(s, gate) > _countTwoQubitGate(
                          seq, gate)):
                    index2seq[i] = [seq]
                else:
                    index2seq[i].append(s)
            start += 1
            if start % 10000 == 0:
                with open(db, 'wb') as f:
                    pickle.dump((start, index2seq), f)
    finally:
        with open(db, 'wb') as f:
            pickle.dump((start, index2seq), f)


# def build():
#     pulses = ['I', 'X', 'Y', 'X/2', 'Y/2', '-X/2', '-Y/2']
#     phases = ['I', 'Z', 'S', '-S']
#     gates = [('CZ', 'CZ'), ('C', 'X'), ('X', 'C'), ('iSWAP', 'iSWAP')]

#     def genSeq(gate):
#         for a in product(pulses, repeat=4):
#             for p1, p2 in product(phases, repeat=2):
#                 yield _short_seq_pair((a[0], gate[0], a[2], p1),
#                                       (a[1], gate[1], a[3], p2))
#         for a in product(pulses, repeat=6):
#             for p1, p2 in product(phases, repeat=2):
#                 yield _short_seq_pair((a[0], gate[0], a[2], gate[0], a[4], p1),
#                                       (a[1], gate[1], a[3], gate[1], a[5], p2))
#         for a in product(pulses, repeat=8):
#             for p1, p2 in product(phases, repeat=2):
#                 yield _short_seq_pair(
#                     (a[0], gate[0], a[2], gate[0], a[4], gate[0], a[6], p1),
#                     (a[1], gate[1], a[3], gate[1], a[5], gate[1], a[7], p2))

#     for gate in gates:
#         for seq in genSeq(gate):
#             seq2index(seq)



import numpy as np


def dBm2Vpp(x, R0=50):
    mP = 10**(x / 10)
    Vrms = np.sqrt(mP * 1e-3 * R0)
    Vpp = 2 * Vrms * np.sqrt(2)
    return Vpp


def Vpp2dBm(x, R0=50):
    Vrms = x / np.sqrt(2) / 2
    mP = Vrms**2 / R0 * 1e3
    dBm = 10 * np.log10(mP)
    return dBm


import operator
from functools import reduce


def cliffordOrder(n: int) -> int:
    """
    Order of complex Clifford group
    
    Sloane, N. J. A. (ed.). "Sequence A003956 (Order of Clifford group)".
    The On-Line Encyclopedia of Integer Sequences. OEIS Foundation.
    https://oeis.org/A003956
    """
    return reduce(operator.mul, (((1 << (2 * j)) - 1) << 2 * j + 1
                                 for j in range(1, n + 1)), 1)



import copy

from waveforms.quantum.circuit.qlisp.config import Config, getConfig, setConfig

from config import config


def test_config():
    cfg = Config.fromdict(config)
    assert isinstance(cfg, Config)
    cfg2 = copy.deepcopy(cfg)
    assert isinstance(cfg2, dict)
    assert not isinstance(cfg2, Config)

    g = cfg.getGate('rfUnitary', 'Q1')
    assert g.amp(3.1) < g.params.amp[1][-1]



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
OPENQASM Lexer.
This is a wrapper around the PLY lexer to support the "include" statement
by creating a stack of lexers.
"""

import os

import numpy as np
import ply.lex as lex

from . import node
from .exceptions import QasmError


class QasmLexer:
    """OPENQASM Lexer.
    This is a wrapper around the PLY lexer to support the "include" statement
    by creating a stack of lexers.
    """

    # pylint: disable=invalid-name,missing-docstring
    # pylint: disable=attribute-defined-outside-init,bad-docstring-quotes

    def __mklexer__(self, filename):
        """Create a PLY lexer."""
        self.lexer = lex.lex(module=self, debug=False)
        self.filename = filename
        self.lineno = 1

        if filename:
            with open(filename) as ifile:
                self.data = ifile.read()
            self.lexer.input(self.data)

    def __init__(self, filename, lib=None):
        """Create the OPENQASM lexer."""
        self.__mklexer__(filename)
        self.stack = []
        self.lib = lib

    def input(self, data):
        """Set the input text data."""
        self.data = data
        self.lexer.input(data)

    def token(self):
        """Return the next token."""
        ret = self.lexer.token()
        return ret

    def pop(self):
        """Pop a PLY lexer off the stack."""
        self.lexer = self.stack.pop()
        self.filename = self.lexer.qasm_file
        self.lineno = self.lexer.qasm_line

    def push(self, filename):
        """Push a PLY lexer on the stack to parse filename."""
        self.lexer.qasm_file = self.filename
        self.lexer.qasm_line = self.lineno
        self.stack.append(self.lexer)
        self.__mklexer__(filename)

    # ---- Beginning of the PLY lexer ----
    literals = r'=()[]{};<>,.+-/*^"'
    reserved = {
        'barrier': 'BARRIER',
        'creg': 'CREG',
        'gate': 'GATE',
        'if': 'IF',
        'measure': 'MEASURE',
        'opaque': 'OPAQUE',
        'qreg': 'QREG',
        'pi': 'PI',
        'reset': 'RESET',
    }
    tokens = [
        'NNINTEGER',
        'REAL',
        'FORMAT',
        'ASSIGN',
        'MATCHES',
        'ID',
        'STRING',
    ] + list(reserved.values())

    def t_REAL(self, t):
        r'(([0-9]+|([0-9]+)?\.[0-9]+|[0-9]+\.)[eE][+-]?[0-9]+)|(([0-9]+)?\.[0-9]+|[0-9]+\.)'
        if np.iscomplex(t):
            return t.real
        else:
            return t

    def t_NNINTEGER(self, t):
        r'[1-9]+[0-9]*|0'
        t.value = int(t.value)
        return t

    def t_ASSIGN(self, t):
        '->'
        return t

    def t_MATCHES(self, t):
        '=='
        return t

    def t_STRING(self, t):
        r'\"([^\\\"]|\\.)*\"'
        return t

    def t_INCLUDE(self, _):
        'include'
        # Now eat up the next two tokens which must be
        # 1 - the name of the include file, and
        # 2 - a terminating semicolon
        #
        # Then push the current lexer onto the stack, create a new one from
        # the include file, and push it onto the stack.
        #
        # When we hit eof (the t_eof) rule, we pop.
        next_token = self.lexer.token()
        lineno = next_token.lineno
        if isinstance(next_token.value, str):
            incfile = next_token.value.strip('"')
        else:
            raise QasmError("Invalid include: must be a quoted string.")

        if self.lib is not None:
            fullpath = self.lib.getQasmLib(incfile)
            if fullpath is not None:
                incfile = fullpath

        next_token = self.lexer.token()
        if next_token is None or next_token.value != ';':
            raise QasmError('Invalid syntax, missing ";" at line', str(lineno))

        if not os.path.exists(incfile):
            raise QasmError(
                'Include file %s cannot be found, line %s, file %s' %
                (incfile, str(next_token.lineno), self.filename))
        self.push(incfile)
        return self.lexer.token()

    def t_FORMAT(self, t):
        r'OPENQASM\s+(\d+)\.(\d+)'
        return t

    def t_COMMENT(self, _):
        r'//.*'
        pass

    def t_ID(self, t):
        r'[a-zA-Z_][a-zA-Z0-9_]*'

        t.type = self.reserved.get(t.value, 'ID')
        if t.type == 'ID':
            t.value = node.Id(t.value, self.lineno, self.filename)
        return t

    def t_newline(self, t):
        r'\n+'
        self.lineno += len(t.value)
        t.lexer.lineno = self.lineno

    def t_eof(self, _):
        if self.stack:
            self.pop()
            return self.lexer.token()
        return None

    t_ignore = ' \t\r'

    def t_error(self, t):
        raise QasmError(
            "Unable to match any token rule, got -->%s<-- "
            "Check your OPENQASM source and any include statements." %
            t.value[0])



from ...qlisp.simulator import seq2mat as _seq2mat
from .mat import normalize


def twoQubitGate(gates):
    return {
        ('CZ', 'CZ'): ('CZ', (0, 1)),
        ('C', 'Z'): ('CZ', (0, 1)),
        ('Z', 'C'): ('CZ', (0, 1)),
        ('CX', 'CX'): ('Cnot', (0, 1)),
        ('XC', 'XC'): ('Cnot', (1, 0)),
        ('CR', 'CR'): ('CR', (0, 1)),
        ('RC', 'RC'): ('CR', (1, 0)),
        ('C', 'X'): ('Cnot', (0, 1)),
        ('X', 'C'): ('Cnot', (1, 0)),
        ('C', 'R'): ('CR', (0, 1)),
        ('R', 'C'): ('CR', (1, 0)),
        ('iSWAP', 'iSWAP'): ('iSWAP', (0, 1)),
        ('SWAP', 'SWAP'): ('SWAP', (0, 1)),
        ('SQiSWAP', 'SQiSWAP'): ('SQiSWAP', (0, 1)),
    }[gates]


def seq2qlisp(seq, qubits):
    if len(seq) > 2:
        raise ValueError("Only support 1 or 2 bits.")
    if len(seq) != len(qubits):
        raise ValueError("seq size and qubit num mismatched.")

    qlisp = []
    for gates in zip(*seq):
        try:
            qlisp.append(twoQubitGate(gates))
        except:
            for gate, i in zip(gates, qubits):
                qlisp.append((gate, i))
    return qlisp


def seq2mat(seq):
    N = len(seq)
    if N > 2:
        raise ValueError("Only support 1 or 2 bits.")
    if N == 1:
        qubits = (1, )
    else:
        qubits = (0, 1)

    return normalize(_seq2mat(seq2qlisp(seq, qubits)))



import asyncio
import time

from . import _kcp


class KCPConnection():

    def __init__(self,
                 address,
                 transport,
                 conv: int = 1234,
                 nodelay: int = 1,
                 interval: int = 20,
                 resend: int = 2,
                 nc: int = 1):
        self._kcp = _kcp.kcp_create(conv, self)
        _kcp.kcp_nodelay(self._kcp, nodelay, interval, resend, nc)
        self.address = address
        self.transport = transport
        self.last_active = time.time()
        self.send_queue = asyncio.Queue()
        self.recv_queue = asyncio.Queue()
        self.fut = asyncio.create_task(self.run())

    def __del__(self):
        try:
            self.fut.cancel()
        except:
            pass
        _kcp.kcp_release(self._kcp)

    def output(self, msg):
        self.transport.sendto(msg, self.address)

    def send(self, buff):
        self.last_active = time.time()
        self.send_queue.put_nowait(buff)

    def _send(self, buff):
        return _kcp.kcp_send(self._kcp, buff)

    async def recv(self):
        return await self.recv_queue.get()

    def recv_nowait(self):
        self.flush()
        try:
            return self.recv_queue.get_nowait()
        except:
            return None

    def _recv(self):
        return _kcp.kcp_recv(self._kcp)

    def _update(self):
        _kcp.kcp_update(self._kcp, self.clock())

    def check(self):
        current = self.clock()
        return _kcp.kcp_check(self._kcp, current) - current

    def clock(self):
        return (time.monotonic_ns() // 1000_000) & 0x7fffffff

    def input(self, buff):
        self.last_active = time.time()
        return _kcp.kcp_input(self._kcp, buff)

    async def update(self):
        while True:
            await asyncio.sleep(self.check() / 1000)
            self._update()

    async def sync(self):
        while True:
            await asyncio.sleep(0.01)
            self.flush()

    async def run(self):
        await asyncio.gather(self.update(), self.sync())

    def flush(self):
        while True:
            buff = self._recv()
            if isinstance(buff, int):
                break
            self.recv_queue.put_nowait(buff)
        while True:
            try:
                buff = self.send_queue.get_nowait()
                if self._send(buff) < 0:
                    self.send_queue.put_nowait(buff)
                    break
            except asyncio.QueueEmpty:
                break

        return True


class KCPProtocol(asyncio.DatagramProtocol):

    def __init__(self):
        self.transport = None
        self.connections = {}

    def connection_made(self, transport):
        self.transport = transport

    def get_connection(self, address):
        if address not in self.connections:
            self.connections[address] = KCPConnection(address, self.transport)
        self.connections[address].last_recv = time.time()
        return self.connections[address]

    def datagram_received(self, datagram, address):
        self.get_connection(address).input(datagram)

    def sendto(self, msg, address):
        self.get_connection(address).send(msg)

    def recvfrom(self):
        for address, conn in list(self.connections.items()):
            buff = conn.recv_nowait()
            if buff:
                return buff, address

    def clean(self):
        current = time.time()
        dead = []
        for address, conn in list(self.connections.items()):
            if current - conn.last_active > 3600:
                dead.append(address)
        for address in dead:
            del self.connections[address]


async def listen(address):
    loop = asyncio.get_running_loop()
    transport, protocol = await loop.create_datagram_endpoint(
        KCPProtocol, local_addr=address)
    return protocol